'''formats/shotcut.py'''

from .utils import get_width_height
from auto_editor.utils.func import aspect_ratio

def frames_to_timecode(frames, fps):
    seconds = frames / fps

    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)

    if(len(str(int(s))) == 1):
        s = '0' + str('{:.3f}'.format(round(s, 3)))
    else:
        s = str('{:.3f}'.format(round(s, 3)))

    return '{:02d}:{:02d}:{}'.format(int(h), int(m), s)

def timecode_to_frames(timecode, fps):
    h, m, s, = timecode.split(':')
    h = int(h)
    m = int(m)
    s = float(s)
    return round((h * 3600 + m * 60 + s) * fps)

def shotcut_xml(inp, temp, output, clips, chunks, fps, log):
    width, height = get_width_height(inp)
    if(width is None or height is None):
        width, height = '1280', '720'
    num, den = aspect_ratio(int(width), int(height))

    global_out = inp.duration

    version = '21.05.18'

    with open(output, 'w', encoding='utf-8') as out:
        out.write('<?xml version="1.0" standalone="no"?>\n')
        out.write('<mlt LC_NUMERIC="C" version="7.1.0" ' +
            'title="Shotcut version {}" '.format(version) +
            'producer="main_bin">\n')
        out.write('\t<profile description="automatic" ' +
            'width="{}" height="{}" '.format(width, height) +
            'progressive="1" sample_aspect_num="1" sample_aspect_den="1" ' +
            'display_aspect_num="{}" display_aspect_den="{}" '.format(num, den) +
            'frame_rate_num="{}" frame_rate_den="1" colorspace="709"/>\n'.format(fps))
        out.write('\t<playlist id="main_bin">\n')
        out.write('\t\t<property name="xml_retain">1</property>\n')
        out.write('\t</playlist>\n')

        # out was the new video length in the original xml
        out.write('\t<producer id="black" in="00:00:00.000" out="{}">\n'.format(global_out))
        out.write('\t\t<property name="length">{}</property>\n'.format(global_out))
        out.write('\t\t<property name="eof">pause</property>\n')
        out.write('\t\t<property name="resource">0</property>\n')
        out.write('\t\t<property name="aspect_ratio">1</property>\n')
        out.write('\t\t<property name="mlt_service">color</property>\n')
        out.write('\t\t<property name="mlt_image_format">rgba</property>\n')
        out.write('\t\t<property name="set.test_audio">0</property>\n')
        out.write('\t</producer>\n')

        out.write('\t<playlist id="background">\n') # same for this out too.
        out.write('\t\t<entry producer="black" in="00:00:00.000" out="{}"/>\n'.format(
            global_out))
        out.write('\t</playlist>\n')

        chains = 0
        producers = 0

        # Speeds like [1.5, 3] don't work because of duration issues, too bad!

        for clip in clips:
            speed = clip[2] / 100
            if(int(speed) == speed):
                speed = int(speed)

            _out = frames_to_timecode(clip[1] / speed, fps)
            length = frames_to_timecode((clip[1] / speed) + 1, fps)

            if(speed == 1):
                resource = inp.path
                caption = inp.basename
                out.write('\t<chain id="chain{}" out="{}">\n'.format(
                    chains, _out))
                chains += 1
            else:
                resource = '{}:{}'.format(speed, inp.path)
                caption = '{} ({}x)'.format(inp.basename, speed)
                out.write('\t<producer id="producer{}" in="00:00:00.000" out="{}">\n'.format(
                    producers, _out))
                producers += 1
                chains += 1 # yes, Shotcut does this.

            out.write('\t\t<property name="length">{}</property>\n'.format(length))
            out.write('\t\t<property name="eof">pause</property>\n')
            out.write('\t\t<property name="resource">{}</property>\n'.format(resource))

            if(speed == 1):
                out.write('\t\t<property name="mlt_service">avformat-novalidate</property>\n')
                out.write('\t\t<property name="seekable">1</property>\n')
                out.write('\t\t<property name="audio_index">1</property>\n')
                out.write('\t\t<property name="video_index">0</property>\n')
                out.write('\t\t<property name="mute_on_pause">0</property>\n')
                out.write('\t\t<property name="shotcut:caption">{}</property>\n'.format(
                    caption))
                out.write('\t\t<property name="xml">was here</property>\n')
            else:
                out.write('\t\t<property name="aspect_ratio">1</property>\n')
                out.write('\t\t<property name="seekable">1</property>\n')
                out.write('\t\t<property name="audio_index">1</property>\n')
                out.write('\t\t<property name="video_index">0</property>\n')
                out.write('\t\t<property name="mute_on_pause">1</property>\n')
                out.write('\t\t<property name="warp_speed">{}</property>\n'.format(speed))
                out.write('\t\t<property name="warp_resource">{}</property>\n'.format(
                    inp.path))
                out.write('\t\t<property name="mlt_service">timewarp</property>\n')
                out.write('\t\t<property name="shotcut:producer">avformat</property>\n')
                out.write('\t\t<property name="video_delay">0</property>\n')
                out.write('\t\t<property name="shotcut:caption">{}</property>\n'.format(caption))
                out.write('\t\t<property name="xml">was here</property>\n')
                out.write('\t\t<property name="warp_pitch">1</property>\n')

            out.write('\t</chain>\n' if speed == 1 else '\t</producer>\n')

        out.write('\t<playlist id="playlist0">\n')
        out.write('\t\t<property name="shotcut:video">1</property>\n')
        out.write('\t\t<property name="shotcut:name">V1</property>\n')

        producers = 0
        for i, clip in enumerate(clips):

            if(speed == 1):
                in_len = clip[0] - 1
            else:
                in_len = max(clip[0] / speed, 0)

            out_len = max((clip[1] - 2) / speed, 0)

            _in = frames_to_timecode(in_len, fps)
            _out = frames_to_timecode(out_len, fps)

            tag_name = 'chain{}'.format(i)
            if(speed != 1):
                tag_name = 'producer{}'.format(producers)
                producers += 1

            out.write('\t\t<entry producer="{}" in="{}" out="{}"/>\n'.format(
                tag_name, _in, _out))

        out.write('\t</playlist>\n')

        out.write('\t<tractor id="tractor0" title="Shotcut version {}" '.format(version) +
            'in="00:00:00.000" out="{}">\n'.format(global_out))
        out.write('\t\t<property name="shotcut">1</property>\n')
        out.write('\t\t<property name="shotcut:projectAudioChannels">2</property>\n')
        out.write('\t\t<property name="shotcut:projectFolder">0</property>\n')
        out.write('\t\t<track producer="background"/>\n')
        out.write('\t\t<track producer="playlist0"/>\n')
        out.write('\t\t<transition id="transition0">\n')
        out.write('\t\t\t<property name="a_track">0</property>\n')
        out.write('\t\t\t<property name="b_track">1</property>\n')
        out.write('\t\t\t<property name="mlt_service">mix</property>\n')
        out.write('\t\t\t<property name="always_active">1</property>\n')
        out.write('\t\t\t<property name="sum">1</property>\n')
        out.write('\t\t</transition>\n')
        out.write('\t\t<transition id="transition1">\n')
        out.write('\t\t\t<property name="a_track">0</property>\n')
        out.write('\t\t\t<property name="b_track">1</property>\n')
        out.write('\t\t\t<property name="version">0.9</property>\n')
        out.write('\t\t\t<property name="mlt_service">frei0r.cairoblend</property>\n')
        out.write('\t\t\t<property name="threads">0</property>\n')
        out.write('\t\t\t<property name="disable">1</property>\n')
        out.write('\t\t</transition>\n')

        out.write('\t</tractor>\n')
        out.write('</mlt>\n')



'''utils/progressbar.py'''

from time import time, localtime
from platform import system
from shutil import get_terminal_size

from .func import get_stdout

def _pretty_time(my_time, ampm):
    # type: (float, bool) -> str
    new_time = localtime(my_time)

    hours = new_time.tm_hour
    minutes = new_time.tm_min

    if(ampm):
        if(hours == 0):
            hours = 12
        if(hours > 12):
            hours -= 12
        ampm = 'PM' if new_time.tm_hour >= 12 else 'AM'
        return '{:02}:{:02} {}'.format(hours, minutes, ampm)
    return '{:02}:{:02}'.format(hours, minutes)

def bar(columns, title, done, togo, percent, new_time):
    bar = '  ⏳{}: [{}{}] {}% done ETA {}'.format(title, done, togo, percent, new_time)
    if(len(bar) > columns - 2):
        bar = bar[:columns - 2]
    else:
        bar += ' ' * (columns - len(bar) - 4)
    try:
        print(bar, end='\r', flush=True)
    except TypeError:
        print(bar, end='\r')


class ProgressBar():
    def tick(self, index):

        if(self.hide):
            return

        percentDone = min(100, round((index+1) / self.total * 100, 1))
        if(percentDone == 0): # Prevent dividing by zero.
            percentPerSec = 0
        else:
            percentPerSec = (time() - self.begin_time) / percentDone

        new_time = _pretty_time(self.begin_time + (percentPerSec * 100), self.ampm)

        if(self.machine):
            index = min(index, self.total)
            raw = int(self.begin_time + (percentPerSec * 100))
            print('{}~{}~{}~{}~{}'.format(
                self.title, index, self.total, self.begin_time, raw),
                end='\r', flush=True)
            return

        columns = get_terminal_size().columns

        if(self.allow_unicode):
            bar_len = max(1, columns - (self.len_title + 50))
            done = round(percentDone / (100 / bar_len))
            togo = '░' * int(bar_len - done)
            bar(columns, self.title, '█' * done, togo, percentDone, new_time)
        else:
            print('   {}% done ETA {}'.format(percentDone, new_time))

    def start(self, total, title='Please wait'):
        self.title = title
        self.len_title = len(title)
        self.total = total
        self.begin_time = time()

        if(self.allow_unicode is None):
            self.allow_unicode = True
            try:
                self.tick(0)
            except UnicodeEncodeError:
                self.allow_unicode = False
                self.tick(0)
        else:
            self.tick(0)

    @staticmethod
    def end():
        print(' ' * max(1, get_terminal_size().columns - 2), end='\r')

    def __init__(self, machine_readable=False, hide=False):
        self.machine = machine_readable
        self.hide = hide
        self.allow_unicode = None

        self.ampm = True
        if(system() == 'Darwin' and not self.machine):
            try:
                date_format = get_stdout(['defaults', 'read',
                    'com.apple.menuextra.clock', 'DateFormat'])
                self.ampm = 'a' in date_format
            except FileNotFoundError:
                pass



'''validate_input.py'''

import os
import re
import sys

from auto_editor.utils.progressbar import ProgressBar

class MyLogger():
    @staticmethod
    def debug(msg):
        print(msg)

    @staticmethod
    def warning(msg):
        print(msg, file=sys.stderr)

    @staticmethod
    def error(msg):
        if("'Connection refused'" in msg):
            pass
        else:
            print(msg, file=sys.stderr)


def parse_bytes(bytestr):
    # Parse a string indicating a byte quantity into an integer.
    matchobj = re.match(r'(?i)^(\d+(?:\.\d+)?)([kMGTPEZY]?)$', bytestr)
    if(matchobj is None):
        return None
    number = float(matchobj.group(1))
    multiplier = 1024.0 ** 'bkmgtpezy'.index(matchobj.group(2).lower())
    return round(number * multiplier)


def sponsor_block_api(_id, categories, log):
    # type: (str, list, Any) -> dict
    from urllib import request
    from urllib.error import HTTPError
    import json

    cat_url = 'categories=['
    for i, cat in enumerate(categories):
        if(i == 0):
            cat_url += '"{}"'.format(cat)
        else:
            cat_url += ',"{}"'.format(cat)
    cat_url += ']'

    try:
        contents = request.urlopen(
            'https://sponsor.ajay.app/api/skipSegments?videoID={}&{}'.format(_id, cat_url))
        return json.loads(contents.read())
    except HTTPError:
        log.warning("Couldn't find skipSegments for id: {}".format(_id))
        return None

def download_video(my_input, args, ffmpeg, log):
    log.conwrite('Downloading video...')
    if('@' in my_input):
        res = my_input[my_input.index('@')+1:]
        if(' ' in res):
            res = res[:res.index(' ')]
        res = res.strip()
        my_input= my_input[:my_input.index(' ')]
    else:
        res = '720'

    outtmpl = re.sub(r'\W+', '-', my_input)
    if(outtmpl.endswith('-mp4')):
        outtmpl = outtmpl[:-4]
    outtmpl += '.mp4'

    if(args.download_dir is not None):
        outtmpl = os.path.join(args.download_dir, outtmpl)

    try:
        import yt_dlp
    except ImportError:
        log.error('Download the yt-dlp python library to download URLs.\n'
            '   pip3 install yt-dlp')

    if(not os.path.isfile(outtmpl)):
        ytbar = ProgressBar(100, 'Downloading')
        def my_hook(d):
            if(d['status'] == 'downloading'):
                ytbar.tick(float(d['_percent_str'].replace('%','')))

        def abspath(path):
            if(path is None):
                return None
            return os.path.abspath(path)

        ydl_opts = {
            'nocheckcertificate': not args.check_certificate,
            'outtmpl': outtmpl,
            'ffmpeg_location': ffmpeg.path,
            'format': f'bestvideo[ext=mp4][height<={res}]+bestaudio[ext=m4a]',
            'ratelimit': parse_bytes(args.limit_rate),
            'logger': MyLogger(),
            'cookiefile': abspath(args.cookies),
            'download_archive': abspath(args.download_archive),
            'progress_hooks': [my_hook],
        }

        for item, key in ydl_opts.items():
            if(item is None):
                del ydl_opts[key]

        try:
            with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                ydl.download([my_input])
        except yt_dlp.utils.DownloadError as error:
            if('format is not available' in str(error)):
                del ydl_opts['format']
                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    ydl.download([my_input])
            else:
                log.error('yt-dlp: Download Error.')

        log.conwrite('')
    return outtmpl


def get_segment(args, my_input, log):
    if(args.block is not None):
        if(args.id is not None):
            return sponsor_block_api(args.id, args.block, log)
        match = re.search(r'youtube\.com/watch\?v=(?P<match>[A-Za-z0-9_-]{11})',
            my_input)
        if(match):
            youtube_id = match.groupdict()['match']
            return sponsor_block_api(youtube_id, args.block, log)
    return None

def valid_input(inputs, ffmpeg, args, log):
    new_inputs = []
    segments = []
    for my_input in inputs:
        if(os.path.isfile(my_input)):
            _, ext = os.path.splitext(my_input)
            if(ext == ''):
                log.error('File must have an extension.')

            new_inputs.append(my_input)
            segments.append(get_segment(args, my_input, log))

        elif(my_input.startswith('http://') or my_input.startswith('https://')):
            new_inputs.append(download_video(my_input, args, ffmpeg, log))
            segments.append(get_segment(args, my_input, log))
        else:
            if(os.path.isdir(my_input)):
                log.error('Input must be a file or url.')
            log.error('Could not find file: {}'.format(my_input))

    return new_inputs, segments



'''audiotsm2/utils/normalizebuffer.py'''

import numpy as np

class NormalizeBuffer():
    """
    A NormalizeBuffer is a mono-channel circular buffer, used to
    normalize audio buffers.
    """
    def __init__(self, length):
        self._data = np.zeros(length)
        self._offset = 0
        self._length = length

    def __repr__(self):
        return "NormalizeBuffer(offset={}, length={}, data=\n{})".format(
            self._offset, self._length, repr(self.to_array()))

    def add(self, window):
        """
        Adds a window element-wise to the NormalizeBuffer.
        """
        n = len(window)
        if n > self._length:
            raise ValueError(
                "the window should be smaller than the NormalizeBuffer")

        # Compute the slice of data where the values will be added
        start = self._offset
        end = self._offset + n

        if end <= self._length:
            self._data[start:end] += window
        else:
            end -= self._length
            self._data[start:] += window[:self._length - start]
            self._data[:end] += window[self._length - start:]

    @property
    def length(self):
        return self._length

    def remove(self, n):
        """Removes the first n values of the NormalizeBuffer."""
        if n >= self._length:
            n = self._length
        if n == 0:
            return

        # Compute the slice of data to reset
        start = self._offset
        end = self._offset + n

        if end <= self._length:
            self._data[start:end] = 0
        else:
            end -= self._length
            self._data[start:] = 0
            self._data[:end] = 0

        self._offset += n
        self._offset %= self._length

    def to_array(self, start=0, end=None):
        """Returns an array containing the same data as the
        NormalizeBuffer, from index start (included) to index
        end (exluded).
        """
        if end is None:
            end = self._length

        start += self._offset
        end += self._offset

        if end <= self._length:
            return np.copy(self._data[start:end])

        end -= self._length
        if start < self._length:
            return np.concatenate((self._data[start:], self._data[:end]))

        start -= self._length
        return np.copy(self._data[start:end])



'''ffwrapper.py'''

# Internal Libraries
import re
import sys
import os.path
import subprocess
from platform import system

# Included Libraries
from auto_editor.utils.func import get_stdout
from auto_editor.utils.log import Log

def regex_match(regex, text):
    match = re.search(regex, text)
    if(match):
        return match.groupdict()['match']
    return None


class File:
    def __repr__(self):
        return str(self.__dict__)

    def __init__(self, ffmpeg, path):
        self.path = path
        self.abspath = os.path.abspath(path)
        self.basename = os.path.basename(path)
        self.dirname = os.path.dirname(os.path.abspath(path))

        f_name, f_ext = os.path.splitext(path)
        self.name = f_name
        self.ext = f_ext

        info = get_stdout([ffmpeg.path, '-hide_banner', '-i', path])

        self.duration = regex_match(r'Duration:\s(?P<match>[\d:.]+),', info)
        self.bitrate = regex_match(r'bitrate:\s(?P<match>\d+\skb\/s)', info)

        self.metadata = {}
        active = False
        active_key = None

        for line in info.split('\n'):
            if(active):
                if(re.search(r'^\s*[A-Z][a-z_]*', line)):
                    break

                key = regex_match(r'^\s*(?P<match>[a-z_]+)', line)
                body = regex_match(r'^\s*[a-z_]*\s*:\s(?P<match>[\w\W]*)', line)

                if(body is not None):
                    if(key is None):
                        if(active_key is not None):
                            self.metadata[active_key] += '\n' + body
                    else:
                        self.metadata[key] = body
                        active_key = key

            if(re.search(r'^\s\sMetadata:', line)):
                active = True

        video_streams = []
        audio_streams = []
        subtitle_streams = []
        fps = None

        sub_exts = {'mov_text': 'srt', 'ass': 'ass', 'webvtt': 'vtt'}

        for line in info.split('\n'):
            if(re.search(r'Stream #', line)):
                s_data = {}
                if(re.search(r'Video:', line)):
                    s_data['width'] = regex_match(r'(?P<match>\d+)x\d+[\s,]', line)
                    s_data['height'] = regex_match(r'\d+x(?P<match>\d+)[\s,]', line)
                    s_data['codec'] = regex_match(r'Video:\s(?P<match>\w+)', line)
                    s_data['bitrate'] = regex_match(r'\s(?P<match>\d+\skb\/s)', line)
                    s_data['fps'] = regex_match(r'\s(?P<match>[\d\.]+)\stbr', line)
                    s_data['lang'] = regex_match(r'Stream #\d+:\d+\((?P<match>\w+)\)', line)

                    if(fps is None):
                        fps = s_data['fps']
                    video_streams.append(s_data)

                elif(re.search(r'Audio:', line)):
                    s_data['codec'] = regex_match(r'Audio:\s(?P<match>\w+)', line)
                    s_data['samplerate'] = regex_match(r'(?P<match>\d+)\sHz', line)
                    s_data['bitrate'] = regex_match(r'\s(?P<match>\d+\skb\/s)', line)
                    s_data['lang'] = regex_match(r'Stream #\d+:\d+\((?P<match>\w+)\)', line)
                    audio_streams.append(s_data)

                elif(re.search(r'Subtitle:', line)):
                    s_data['lang'] = regex_match(r'Stream #\d+:\d+\((?P<match>\w+)\)', line)
                    s_data['codec'] = regex_match(r'Subtitle:\s(?P<match>\w+)', line)
                    s_data['ext'] = sub_exts.get(s_data['codec'], 'vtt')
                    subtitle_streams.append(s_data)

        self.fps = fps
        self.video_streams = video_streams
        self.audio_streams = audio_streams
        self.subtitle_streams = subtitle_streams


class FFmpeg():

    @staticmethod
    def _set_ff_path(ff_location, my_ffmpeg):
        # type: (str | None, bool) -> str
        if(ff_location is not None):
            return ff_location
        if(my_ffmpeg or system() not in ['Windows', 'Darwin']):
            return 'ffmpeg'
        program = 'ffmpeg' if system() == 'Darwin' else 'ffmpeg.exe'
        dirpath = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(dirpath, 'ffmpeg', system(), program)

    def __init__(self, ff_location=None, my_ffmpeg=False, debug=False):
        self.debug = debug
        self.path = self._set_ff_path(ff_location, my_ffmpeg)
        try:
            _version = get_stdout([self.path, '-version']).split('\n')[0]
            _version = _version.replace('ffmpeg version', '').strip()
            self.version = _version.split(' ')[0]
        except FileNotFoundError:
            if(system() == 'Darwin'):
                Log().error('No ffmpeg found, download via homebrew or restore the '
                    'included binary.')
            if(system() == 'Windows'):
                Log().error('No ffmpeg found, download ffmpeg with your favorite package '
                    'manager (ex chocolatey), or restore the included binary.')

            Log().error('ffmpeg must be installed and on PATH.')

    def print(self, message):
        if(self.debug):
            print('FFmpeg: {}'.format(message), file=sys.stderr)

    def print_cmd(self, cmd):
        if(self.debug):
            print('FFmpeg run: {}\n'.format(' '.join(cmd)), file=sys.stderr)

    def run(self, cmd):
        cmd = [self.path, '-y', '-hide_banner'] + cmd
        if(not self.debug):
            cmd.extend(['-nostats', '-loglevel', 'error'])
        self.print_cmd(cmd)
        subprocess.call(cmd)

    def run_check_errors(self, cmd, log, show_out=False):
        process = self.Popen(cmd, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        stdout, stderr = process.communicate()
        process.stdin.close()
        output = stderr.decode()

        error_list = [
            r"Unknown encoder '.*'",
            r"-q:v qscale not available for encoder\. Use -b:v bitrate instead\.",
            r'Specified sample rate .* is not supported',
            r'Unable to parse option value ".*"',
            r'Error setting option .* to value .*\.',
            r"Undefined constant or missing '.*' in '.*'",
        ]

        if(self.debug):
            print('stderr: {}'.format(stderr.decode()))
            print('stdout: {}'.format(stdout.decode()))

        for item in error_list:
            check = re.search(item, output)
            if(check):
                log.error(check.group())

        if(show_out and not self.debug):
            print('stderr: {}'.format(stderr.decode()))
            print('stdout: {}'.format(stdout.decode()))

    def file_info(self, path):
        return File(self, path)

    def Popen(self, cmd, stdin=None, stdout=subprocess.PIPE, stderr=None):
        cmd = [self.path] + cmd
        self.print_cmd(cmd)
        return subprocess.Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr)

    def pipe(self, cmd):
        # type: (list[str]) -> str
        cmd = [self.path, '-y'] + cmd

        self.print_cmd(cmd)
        output = get_stdout(cmd)
        self.print(output)
        return output



'''utils/types.py'''

import os.path
import sys

from .func import clean_list

def error(message):
    # type: (str) -> None
    print('Error! {}'.format(message), file=sys.stderr)
    sys.exit(1)

def split_num_str(val, error_func):
    # type: (str, Any) -> tuple(2)
    index = 0
    for item in val:
        if(not item.isdigit() and item != ' ' and item != '.' and item != '-'):
            break
        index += 1
    num, unit = val[:index], val[index:]
    if('.' in num):
        try:
            float(num)
        except ValueError:
            error_func('{} is not a valid number')
        return float(num), unit
    if('-' in num):
        try:
            int(num)
        except ValueError:
            error_func('{} is not a valid number')
    return int(num), unit

def file_type(path):
    # type: (str) -> str
    if(not os.path.isfile(path)):
        error('Auto-Editor could not find the file: {}'.format(path))
    return path

def unit_check(unit, allowed_units):
    # type: (list, str) -> None
    if(unit not in allowed_units):
        error('Unsupported unit: {}'.format(unit))


def float_type(val):
    # type: (str) -> float
    num, unit = split_num_str(val, error)
    unit_check(unit, ['%', ''])
    if(unit == '%'):
        return float(num / 100)
    return float(num)

def sample_rate_type(val):
    # type: (str) -> int
    num, unit = split_num_str(val, error)
    unit_check(unit, ['Hz', 'kHz', ''])
    if(unit == 'kHz'):
        return int(num * 1000)
    return int(num)

def frame_units():
    return ['f', 'frame', 'frames']

def second_units():
    return ['s', 'sec', 'secs', 'second', 'seconds']

def frame_type(val):
    # type: (str) -> int | str
    num, unit = split_num_str(val, error)
    unit_check(unit, [''] + frame_units() + second_units())

    if(unit in second_units()):
        return str(num).strip()
    return int(num)

def comma_type(inp, min_args=1, max_args=None, name=''):
    inp = clean_list(inp.split(','), '\r\n\t')
    if(min_args > len(inp)):
        error('Too few comma arguments for {}.'.format(name))
    if(max_args is not None and len(inp) > max_args):
        error('Too many comma arguments for {}.'.format(name))
    return inp

def range_type(inp):
    return comma_type(inp, 2, 2, 'range_type')

def speed_range_type(inp):
    return comma_type(inp, 3, 3, 'speed_range_type')

def block_type(inp):
    return comma_type(inp, 1, None, 'block_type')



'''utils/container.py'''

pcm_formats = [
    'pcm_s16le', # default format

    'pcm_alaw',
    'pcm_f32be',
    'pcm_f32le',
    'pcm_f64be',
    'pcm_f64le',
    'pcm_mulaw',
    'pcm_s16be',
    'pcm_s24be',
    'pcm_s24le',
    'pcm_s32be',
    'pcm_s32le',
    'pcm_s8',
    'pcm_u16be',
    'pcm_u16le',
    'pcm_u24be',
    'pcm_u24le',
    'pcm_u32be',
    'pcm_u32le',
    'pcm_u8',
    'pcm_vidc',
]

# Define aliases
h265 = {
    'name': 'H.265 / High Efficiency Video Coding (HEVC) / MPEG-H Part 2',
    'allow_video': True,
    'vcodecs': ['hevc', 'mpeg4', 'h264'],
}
h264 = {
    'name': 'H.264 / Advanced Video Coding (AVC) / MPEG-4 Part 10',
    'allow_video': True,
    'vcodecs': ['h264', 'mpeg4', 'hevc'],
}
aac = {
    'name': 'Advanced Audio Coding',
    'allow_audio': True,
    'max_audio_streams': 1,
    'acodecs': ['aac'],
    'astrict': True,
}
ass = {
    'name': 'SubStation Alpha',
    'allow_subtitle': True,
    'scodecs': ['ass', 'ssa'],
    'max_subtitle_streams': 1,
    'sstrict': True,
}
mp4 = {
    'name': 'MP4 / MPEG-4 Part 14',
    'allow_video': True,
    'allow_audio': True,
    'allow_subtitle': True,
    'vcodecs': ['h264', 'hevc', 'vp9', 'av1', 'mpeg4', 'mpeg2video', 'mjpeg'],
    'acodecs': ['aac', 'mp3', 'opus', 'flac', 'vorbis', 'mp2'],
    'disallow_v': ['prores', 'apng', 'gif', 'msmpeg4v3', 'flv1', 'vp8', 'rawvideo'],
    'disallow_a': pcm_formats,
}
ogg = {
    'allow_video': True,
    'allow_audio': True,
    'allow_subtitle': True,
    'vcodecs': ['libtheora', 'theora'],
    'acodecs': ['libvorbis', 'vorbis', 'flac', 'opus', 'speex'],
    'vstrict': True,
    'astrict': True,
},

containers = {

    # Aliases section

    'aac': aac,
    'adts': aac,
    'ass': ass,
    'ssa': ass,
    '264': h264,
    'h264': h264,
    '265': h265,
    'h265': h265,
    'hevc': h265,

    'mp4': mp4,
    'm4a': mp4,

    'ogg': ogg,
    'ogv': ogg,

    'apng': {
        'name': 'Animated Portable Network Graphics',
        'allow_video': True,
        'max_video_streams': 1,
        'vcodecs': ['apng'],
        'vstrict': True,
    },
    'gif': {
        'name': 'Graphics Interchange Format',
        'allow_video': True,
        'max_video_streams': 1,
        'vcodecs': ['gif'],
        'vstrict': True,
    },
    'wav': {
        'name': 'Waveform Audio File Format',
        'allow_audio': True,
        'max_audio_streams': 1,
        'acodecs': pcm_formats + ['mp3'],
        'astrict': True,
    },
    'ast': {
        'name': 'AST / Audio Stream',
        'allow_audio': True,
        'acodecs': ['pcm_s16be_planar'],
    },
    'mp3': {
        'name': 'MP3 / MPEG-2 Audio Layer 3',
        'allow_audio': True,
        'max_audio_streams': 1,
        'acodecs': ['mp3'],
        'astrict': True,
    },
    'opus': {
        'name': 'Opus',
        'allow_audio': True,
        'acodecs': ['opus', 'flac', 'libvorbis', 'vorbis', 'speex'],
        'astrict': True,
    },
    'oga': {
        'allow_audio': True,
        'acodecs': ['flac', 'libvorbis', 'vorbis', 'opus', 'speex'],
        'astrict': True,
    },
    'flac': {
        'name': 'Free Lossless Audio Codec',
        'allow_audio': True,
        'max_audio_streams': 1,
        'acodecs': ['flac'],
    },
    'webm': {
        'name': 'WebM',
        'allow_video': True,
        'allow_audio': True,
        'allow_subtitle': True,
        'vcodecs': ['vp9', 'vp8', 'av1', 'libaom-av1'],
        'acodecs': ['opus', 'vorbis'],
        'scodecs': ['webvtt'],
        'vstrict': True,
        'astrict': True,
        'sstrict': True,
    },
    'srt': {
        'name': 'SubRip Text / Subtitle Resource Tracks',
        'allow_subtitle': True,
        'scodecs': ['srt'],
        'max_subtitle_streams': 1,
        'sstrict': True,
    },
    'vtt': {
        'name': 'Web Video Text Track',
        'allow_subtitle': True,
        'scodecs': ['webvtt'],
        'max_subtitle_streams': 1,
        'sstrict': True,
    },
    'avi': {
        'name': 'Audio Video Interleave',
        'allow_video': True,
        'allow_audio': True,
        'vcodecs': ['mpeg4', 'h264', 'prores', 'mjpeg', 'mpeg2video', 'rawvideo'],
        'acodecs': ['mp3', 'aac', 'vorbis', 'mp2'],
        'disallow_v': ['hevc', 'apng', 'gif'],
    },
    'wmv': {
        'name': 'Windows Media Video',
        'allow_video': True,
        'allow_audio': True,
        'vcodecs': ['msmpeg4v3', 'h264', 'mpeg4', 'mpeg2video', 'mjpeg', 'rawvideo'],
        'acodecs': ['wmav2', 'aac', 'flac'],
        'disallow_v': ['prores', 'hevc', 'apng', 'gif'],
    },
    'mkv': {
        'name': 'Matroska',
        'allow_video': True,
        'allow_audio': True,
        'allow_subtitle': True,
        'vcodecs': ['h264', 'hevc', 'vp9', 'vp8', 'prores', 'mpeg4', 'mpeg2video',
            'msmpeg4v3', 'mjpeg', 'gif', 'rawvideo'],
        'acodecs': ['libvorbis', 'vorbis', 'opus', 'flac', 'aac', 'mp2'],
        'disallow_v': ['apng'],
    },
    'mka': {
        'name': 'Matroska Audio',
        'allow_audio': True,
        'acodecs': ['libvorbis', 'vorbis', 'opus', 'flac', 'aac', 'mp2'],
    },
    'mov': {
        'name': 'QuickTime / MOV',
        'allow_video': True,
        'allow_audio': True,
        'allow_subtitle': True,
        'vcodecs': ['h264', 'hevc', 'prores', 'mpeg4', 'mpeg2video', 'msmpeg4v3', 'mjpeg',
            'gif', 'flv1', 'rawvideo'],
        'acodecs': ['aac', 'mp3', 'mp2', 'vorbis'],
        'disallow_a': ['opus', 'flac'],
        'disallow_v': ['apng', 'vp9', 'vp8'],
    },
    'swf': {
        'name': 'ShockWave Flash / Small Web Format',
        'allow_video': True,
        'allow_audio': True,
        'vcodecs': ['flv1', 'mjpeg'],
        'acodecs': ['mp3'],
        'vstrict': True,
        'astrict': True,
        'samplerate': [44100, 22050, 11025],
    },
    'not_in_here': {
        'allow_video': True,
        'allow_audio': True,
        'allow_subtitle': True,
    },
    'default': {
        'name': None,
        'allow_video': False,
        'allow_audio': False,
        'allow_subtitle': False,
        'max_video_streams': None,
        'max_audio_streams': None,
        'max_subtitle_streams': None,
        'vcodecs': None,
        'acodecs': None,
        'scodecs': None,
        'vstrict': False,
        'astrict': False,
        'sstrict': False,
        'disallow_v': [],
        'disallow_a': [],
        'samplerate': None, # Any samplerate is allowed.
    },
}

def get_rules(key):
    # type: (str) -> dict
    rules = containers['default']
    if(key in containers):
        rules.update(containers[key])
    else:
        rules.update(containers['not_in_here'])
    return rules

if(__name__ == '__main__'):
    import json
    import subprocess

    from auto_editor.ffwrapper import FFmpeg

    with open('../../container.json', 'w') as outfile:
        json.dump(containers, outfile)

    ffmpeg = FFmpeg()

    out = '../../out.opus'
    ffmpeg.run(['-i', '../../example.mp4', '-c:a', 'flac', '-strict', '-2', out])
    subprocess.call(['aeinfo', out])



'''subcommands/desc.py'''

import sys

def desc_options(parser):
    parser.add_argument('--ffmpeg_location', default=None,
        help='Point to your custom ffmpeg file.')
    parser.add_argument('--help', '-h', action='store_true',
        help='Print info about the program or an option and exit.')
    parser.add_argument('input', nargs='*',
        help='The path to file(s)')
    return parser

def main(sys_args=sys.argv[1:]):
    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.log import Log
    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('desc', auto_editor.version,
        description="Print the video's metadata description.")
    parser = desc_options(parser)

    log = Log()

    try:
        args = parser.parse_args(sys_args)
    except vanparse.ParserError as e:
        log.error(str(e))

    ffmpeg = FFmpeg(args.ffmpeg_location, debug=False)

    print('')
    for input_file in args.input:
        inp = ffmpeg.file_info(input_file)
        if('description' in inp.metadata):
            print(inp.metadata['description'], end='\n\n')
        else:
            print('No description.', end='\n\n')

    log.cleanup()


if(__name__ == '__main__'):
    main()



'''audiotsm2/phasevocoder.py'''

import numpy as np

from auto_editor.audiotsm2.base import AnalysisSynthesisTSM
from auto_editor.audiotsm2.utils.windows import hanning


def find_peaks(amplitude):
    """
    A value is considered to be a peak if it is higher than its four closest
    neighbours.
    """

    # Pad the array with -1 at the beginning and the end to avoid overflows.
    padded = np.concatenate((-np.ones(2), amplitude, -np.ones(2)))

    # Shift the array by one/two values to the left/right
    shifted_l2 = padded[:-4]
    shifted_l1 = padded[1:-3]
    shifted_r1 = padded[3:-1]
    shifted_r2 = padded[4:]

    # Compare the original array with the shifted versions.
    peaks = ((amplitude >= shifted_l2) & (amplitude >= shifted_l1) &
             (amplitude >= shifted_r1) & (amplitude >= shifted_r2))

    return peaks


def all_peaks(amplitude):
    """
    A peak finder that considers all values to be peaks.
    This is used for the phase vocoder without phase locking.
    """
    return np.ones_like(amplitude, dtype=bool)


def get_closest_peaks(peaks):
    """
    Returns an array containing the index of the closest peak of each index.
    """
    closest_peak = np.empty_like(peaks, dtype=int)
    previous = -1
    for i, is_peak in enumerate(peaks):
        if is_peak:
            if previous >= 0:
                closest_peak[previous:(previous + i) // 2 + 1] = previous
                closest_peak[(previous + i) // 2 + 1:i] = i
            else:
                closest_peak[:i] = i
            previous = i
    closest_peak[previous:] = previous

    return closest_peak


class PhaseVocoderConverter():
    """
    A Converter implementing the phase vocoder time-scale modification procedure.
    """

    def __init__(self, channels, frame_length, analysis_hop, synthesis_hop, peak_finder):
        self._channels = channels
        self._frame_length = frame_length
        self._synthesis_hop = synthesis_hop
        self._analysis_hop = analysis_hop
        self._find_peaks = peak_finder

        # Centers of the FFT frequency bins
        self._center_frequency = np.fft.rfftfreq(frame_length) * 2 * np.pi
        fft_length = len(self._center_frequency)

        self._first = True

        self._previous_phase = np.empty((channels, fft_length))
        self._output_phase = np.empty((channels, fft_length))

        # Buffer used to compute the phase increment and the instantaneous frequency
        self._buffer = np.empty(fft_length)

    def clear(self):
        self._first = True

    def convert_frame(self, frame):
        for k in range(0, self._channels):
            # Compute the FFT of the analysis frame
            stft = np.fft.rfft(frame[k])
            amplitude = np.abs(stft)
            phase = np.angle(stft)
            del stft

            peaks = self._find_peaks(amplitude)
            closest_peak = get_closest_peaks(peaks)

            if self._first:
                # Leave the first frame unchanged
                self._output_phase[k, :] = phase
            else:
                # Compute the phase increment
                self._buffer[peaks] = (
                    phase[peaks] - self._previous_phase[k, peaks] -
                    self._analysis_hop * self._center_frequency[peaks]
                )

                # Unwrap the phase increment
                self._buffer[peaks] += np.pi
                self._buffer[peaks] %= 2 * np.pi
                self._buffer[peaks] -= np.pi

                # Compute the instantaneous frequency (in the same buffer,
                # since the phase increment wont be required after that)
                self._buffer[peaks] /= self._analysis_hop
                self._buffer[peaks] += self._center_frequency[peaks]

                self._buffer[peaks] *= self._synthesis_hop
                self._output_phase[k][peaks] += self._buffer[peaks]

                # Phase locking
                self._output_phase[k] = (
                    self._output_phase[k][closest_peak] +
                    phase - phase[closest_peak]
                )

                # Compute the new stft
                output_stft = amplitude * np.exp(1j * self._output_phase[k])

                frame[k, :] = np.fft.irfft(output_stft).real

            # Save the phase for the next analysis frame
            self._previous_phase[k, :] = phase
            del phase
            del amplitude

        self._first = False

        return frame

    def set_analysis_hop(self, analysis_hop):
        self._analysis_hop = analysis_hop


class PhaseLocking():
    NONE = 0
    IDENTITY = 1

    @classmethod
    def from_str(cls, name):
        if name.lower() == 'none':
            return cls.NONE
        elif name.lower() == 'identity':
            return cls.IDENTITY
        else:
            raise ValueError('Invalid phase locking name: "{}"'.format(name))


def phasevocoder(channels, speed=1., frame_length=2048, analysis_hop=None,
    synthesis_hop=None, phase_locking=PhaseLocking.IDENTITY):

    if synthesis_hop is None:
        synthesis_hop = frame_length // 4

    if analysis_hop is None:
        analysis_hop = int(synthesis_hop * speed)

    analysis_window = hanning(frame_length)
    synthesis_window = hanning(frame_length)

    if phase_locking == PhaseLocking.NONE:
        peak_finder = all_peaks
    elif phase_locking == PhaseLocking.IDENTITY:
        peak_finder = find_peaks
    else:
        raise ValueError('Invalid phase_locking value: "{}"'.format(phase_locking))

    converter = PhaseVocoderConverter(channels, frame_length, analysis_hop,
        synthesis_hop, peak_finder)

    return AnalysisSynthesisTSM(converter, channels, frame_length, analysis_hop,
        synthesis_hop, analysis_window, synthesis_window)



#!/usr/bin/env python3
'''__main__.py'''

# Internal Libraries
import os
import sys
import tempfile

# Included Libraries
import auto_editor
import auto_editor.vanparse as vanparse
import auto_editor.utils.func as usefulfunctions

from auto_editor.utils.progressbar import ProgressBar
from auto_editor.utils.func import set_output_name
from auto_editor.utils.log import Log, Timer
from auto_editor.ffwrapper import FFmpeg
from auto_editor.edit import edit_media

def main_options(parser):
    from auto_editor.utils.types import (file_type, float_type, sample_rate_type,
        frame_type, range_type, speed_range_type, block_type)

    parser.add_text('Progress Options')
    parser.add_argument('--machine_readable_progress', action='store_true',
        help='Set progress bar that is easier to parse.')
    parser.add_argument('--no_progress', action='store_true',
        help='Do not display any progress at all.')

    parser.add_text('Object Options')

    basic_obj_keywords = [
        {'start': (int, ''),},
        {'dur': (int, ''),},
        {'x1': (int, ''),},
        {'y1': (int, ''),},
        {'x2': (int, ''),},
        {'y2': (int, ''),},
        {'fill': (str, '#000'),},
        {'width': (int, 0),},
        {'outline': (str,'blue'),},
    ]

    rect = basic_obj_keywords[:]
    rect.append({'type': 'rectangle'})

    ellipse = basic_obj_keywords[:]
    ellipse.append({'type': 'ellipse'})

    parser.add_argument('--add_text', nargs='*', type=dict,
        keywords=[
            {'start': (int, ''),},
            {'dur': (int, ''),},
            {'content': (str, ''),},
            {'x': (int, 'centerX'),},
            {'y': (int, 'centerY'),},
            {'size': (int, 30),},
            {'font': (str, 'default'),},
            {'align': (str, 'left'),},
            {'fill': (str, '#FFF'),},
            {'type': 'text',},
        ],
        help='Add a text object to the timeline.'
    )
    parser.add_argument('--add_rectangle', nargs='*', type=dict,
        keywords=rect,
        help='Add a rectangle object to the timeline.')
    parser.add_argument('--add_ellipse', nargs='*', type=dict,
        keywords=ellipse,
        help='Add an ellipse object to the timeline.',
        manual='The x and y coordinates specify a bounding box where the ellipse is '\
            'drawn.')

    # parser.add_argument('--add_image', nargs='*', type=dict,
    #     keywords=[


    #     ],
    #     help='Add an image object onto the timeline.')

    parser.add_text('URL Download Options')
    parser.add_argument('--download_dir', type=str, default=None,
        help='The directory where the downloaded file is placed.')
    parser.add_argument('--limit_rate', '-rate', default='3m',
        help='The maximum download rate in bytes per second (50k, 4.2m)')
    parser.add_argument('--id', type=str, default=None,
        help='Manually set the YouTube ID the video belongs to.')
    parser.add_argument('--block', type=block_type,
        help='Mark all sponsors sections as silent.',
        manual='Only for YouTube URLs. This uses the SponsorBlock API.\n'
            'Choices can include: sponsor intro outro selfpromo interaction music_offtopic')

    parser.add_argument('--download_archive', type=file_type, default=None,
        help='Download only videos not listed in archive file. Record the IDs of'
             ' all downloads.')
    parser.add_argument('--cookies', type=file_type, default=None,
        help='The file to read cookies from and dump the cookie jar in.')
    parser.add_argument('--check_certificate', action='store_true',
        help='Check the website certificate before downloading.')

    parser.add_text('Motion Detection Options')
    parser.add_argument('--motion_threshold', type=float_type, default=0.02,
        range='0 to 1',
        help='How much motion is required to be considered "moving"')
    parser.add_blank()
    parser.add_argument('--md_dilates', type=int, default=2, range='0 to 5',
        help='Set how many times a frame is dilated before being compared.')
    parser.add_argument('--md_width', type=int, default=400, range='1 to Infinity',
        help="Scale the frame to this width before being compared.")
    parser.add_argument('--md_blur', type=int, default=21, range='0 to Infinity',
        help='Set the strength of the blur applied to a frame before being compared.')

    parser.add_text('Exporting as Media Options')
    parser.add_argument('--video_bitrate', '-b:v', default='10m',
        help='Set the number of bits per second for video.')
    parser.add_argument('--audio_bitrate', '-b:a', default='unset',
        help='Set the number of bits per second for audio.')
    parser.add_argument('--sample_rate', '-ar', type=sample_rate_type,
        help='Set the sample rate of the input and output videos.')
    parser.add_argument('--video_codec', '-vcodec', '-c:v', default='auto',
        help='Set the video codec for the output media file.')
    parser.add_argument('--audio_codec', '-acodec', '-c:a', default='auto',
        help='Set the audio codec for the output media file.')
    parser.add_argument('--preset', '-preset', default='unset',
        choices=['ultrafast', 'superfast', 'veryfast', 'faster', 'fast', 'medium',
            'slow', 'slower', 'veryslow', 'unset'],
        help='Set the preset for ffmpeg to help save file size or increase quality.')
    parser.add_argument('--tune', '-tune', default='unset',
        choices=['film', 'animation', 'grain', 'stillimage', 'fastdecode',
            'zerolatency', 'none', 'unset'],
        help='Set the tune for ffmpeg to compress video better in certain circumstances.')
    parser.add_argument('--constant_rate_factor', '-crf', default='unset', range='0 to 51',
        help='Set the quality for video using the crf method.')
    parser.add_argument('--video_quality_scale', '-qscale:v', '-q:v', default='unset',
        range='1 to 31',
        help='Set a value to the ffmpeg option -qscale:v')
    parser.add_argument('--scale', type=float_type, default=1,
        help='Scale the output media file by a certain factor.')
    parser.add_argument('--has_vfr', default='unset',
        choices=['unset', 'yes', 'no'],
        help='Set whether the input file(s) have Variable Frame Rates',
        manual='Detecting Variable Frame Rates takes a lot of time, skip variable frame rate scan, saving time for big video files.')

    parser.add_text('Miscellaneous Options')
    parser.add_argument('--background', type=str, default='#000',
        help='Set the color of the background that is visible when the video is moved.')
    parser.add_argument('--combine_files', action='store_true',
        help='Combine all input files into one before editing.')

    parser.add_text('Manual Editing Options')
    parser.add_argument('--cut_out', type=range_type, nargs='*',
        help='The range of media that will be removed completely, regardless of the '
            'value of silent speed.')
    parser.add_argument('--add_in', type=range_type, nargs='*',
        help='The range of media that will be added in, opposite of --cut_out')
    parser.add_blank()
    parser.add_argument('--mark_as_loud', type=range_type, nargs='*',
        help='The range that will be marked as "loud".')
    parser.add_argument('--mark_as_silent', type=range_type, nargs='*',
        help='The range that will be marked as "silent".')
    parser.add_argument('--set_speed_for_range', type=speed_range_type, nargs='*',
        help='Set an arbitrary speed for a given range.',
        manual='This option takes 3 arguments delimited with commas and they are as follows:\n'
            ' Speed\n'
            ' - How fast the media plays. Speeds 0 or below and 99999 or above will be cut completely.\n'
            ' Start\n'
            ' - When the speed first gets applied. The default unit is in frames, but second units can also be used.\n'
            ' End\n'
            ' - When the speed stops being applied. It can use both frame and second units.\n')

    parser.add_text('Select Editing Source Options')
    parser.add_argument('--edit_based_on', '--edit', default='audio',
        choices=['audio', 'motion', 'none', 'all', 'not_audio', 'not_motion',
            'audio_or_motion', 'audio_and_motion', 'audio_xor_motion',
            'audio_and_not_motion', 'not_audio_and_motion', 'not_audio_and_not_motion'],
        help='Decide which method to use when making edits.')
    parser.add_argument('--keep_tracks_seperate', action='store_true',
        help="Don't combine audio tracks when exporting.")
    parser.add_blank()
    parser.add_argument('--cut_by_this_audio', '-ca', type=file_type,
        help="Base cuts by this audio file instead of the video's audio.")
    parser.add_argument('--cut_by_this_track', '-ct', type=int, default=0,
        range='0 to the number of audio tracks minus one',
        help='Base cuts by a different audio track in the video.')
    parser.add_argument('--cut_by_all_tracks', '-cat', action='store_true',
        help='Combine all audio tracks into one before basing cuts.')

    parser.add_text('Export Mode Options')
    parser.add_argument('--export_to_premiere', '-exp', action='store_true',
        help='Export as an XML file for Adobe Premiere Pro instead of making a media file.')
    parser.add_argument('--export_to_final_cut_pro', '-exf', action='store_true',
        help='Export as an XML file for Final Cut Pro instead of making a media file.')
    parser.add_argument('--export_to_shotcut', '-exs', action='store_true',
        help='Export as an XML timeline file for Shotcut instead of making a media file.')
    parser.add_argument('--export_as_json', action='store_true',
        help='Export as a JSON file that can be read by auto-editor later.')
    parser.add_argument('--export_as_audio', '-exa', action='store_true',
        help='Export as a WAV audio file.')
    parser.add_argument('--export_as_clip_sequence', '-excs', action='store_true',
        help='Export as multiple numbered media files.')

    parser.add_text('Utility Options')
    parser.add_argument('--no_open', action='store_true',
        help='Do not open the file after editing is done.')
    parser.add_argument('--temp_dir', default=None,
        help='Set where the temporary directory is located.',
        manual='If not set, tempdir will be set with Python\'s tempfile module\n'
            'The directory doesn\'t have to exist beforehand, however, the root path must be valid.'
            'For Windows users, this file will be in the C drive.\n'
            'The temp file can get quite big if you\'re generating a huge video, so '
            'make sure your location has enough space.')
    parser.add_argument('--ffmpeg_location', default=None,
        help='Set a custom path to the ffmpeg location.',
        manual='This takes precedence over `--my_ffmpeg`.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='Use the ffmpeg on your PATH instead of the one packaged.',
        manual='This is equivalent to `--ffmpeg_location ffmpeg`.')

    parser.add_text('Display Options')
    parser.add_argument('--version', action='store_true',
        help="Display the program's version and halt.")
    parser.add_argument('--debug', action='store_true',
        help='Show debugging messages and values.')
    parser.add_argument('--show_ffmpeg_debug', action='store_true',
        help='Show ffmpeg progress and output.')
    parser.add_argument('--quiet', '-q', action='store_true',
        help='Display less output.')
    parser.add_argument('--preview', action='store_true',
        help='Show stats on how the input will be cut and halt.')

    parser.add_text('Editing Options')

    parser.add_argument('--silent_threshold', '-t', type=float_type, default=0.04,
        range='0 to 1',
        help='Set the volume that frames audio needs to surpass to be "loud".',
        manual='Silent threshold is a percentage where 0% represents absolute silence and '
            '100% represents the highest volume in the media file.\n'
            'Setting the threshold to `0%` will cut only out areas where area is '
            'absolutely silence while a value of 4% will cut ')
    parser.add_argument('--frame_margin', '--margin', '-m', type=frame_type, default=6,
        range='0 to Infinity',
        help='Set how many "silent" frames on either side on the "loud" sections to be '
            'included.')
    parser.add_argument('--silent_speed', '-s', type=float_type, default=99999,
        range='Any number. Values <= 0 or >= 99999 will be cut out.',
        help='Set the speed that "silent" sections should be played at.')
    parser.add_argument('--video_speed', '--sounded_speed', '-v', type=float_type,
        default=1.00,
        range='Any number. Values <= 0 or >= 99999 will be cut out.',
        help='Set the speed that "loud" sections should be played at.')
    parser.add_argument('--min_clip_length', '-mclip', type=frame_type, default=3,
        range='0 to Infinity',
        help='Set the minimum length a clip can be. If a clip is too short, cut it.')
    parser.add_argument('--min_cut_length', '-mcut', type=frame_type, default=6,
        range='0 to Infinity',
        help="Set the minimum length a cut can be. If a cut is too short, don't cut")

    parser.add_blank()
    parser.add_argument('--help', '-h', action='store_true',
        help='Print info about the program or an option and exit.')

    parser.add_argument('--output_file', '--output', '-o', nargs='*',
        help='Set the name(s) of the new output.')
    parser.add_argument('input', nargs='*',
        help='File(s) or URL(s) that will be edited.')

    parser.add_blank()
    parser.add_text('  Have an issue? Make an issue. Visit '
            'https://github.com/wyattblue/auto-editor/issues\n\n  The help option '
            'can also be used on a specific option:\n     auto-editor '
            '--frame_margin --help\n')
    return parser


def main():
    desc = ('Auto-Editor is an automatic video/audio creator and editor. '
        'By default, it will detect silence and create a new video with those '
        'sections cut out. By changing some of the options, you can export to a '
        'traditional editor like Premiere Pro and adjust the edits there, adjust '
        'the pacing of the cuts, and change the method of editing like using audio '
        'loudness and video motion to judge making cuts.\nRun:\n    auto-editor '
        '--help\n\nTo get the list of options.\n')

    parser = vanparse.ArgumentParser('Auto-Editor', auto_editor.version, description=desc)

    subcommands = ['create', 'test', 'info', 'levels', 'grep', 'subdump', 'desc']

    if(len(sys.argv) > 1 and sys.argv[1] in subcommands):
        obj = __import__('auto_editor.subcommands.{}'.format(sys.argv[1]),
            fromlist=['subcommands'])
        obj.main(sys.argv[2:])
        sys.exit()
    else:
        parser = main_options(parser)
        try:
            args = parser.parse_args(sys.argv[1:])
        except vanparse.ParserError as e:
            Log().error(str(e))

    timer = Timer(args.quiet)

    exporting_to_editor = (args.export_to_premiere or args.export_to_final_cut_pro or
        args.export_to_shotcut)
    making_data_file = exporting_to_editor or args.export_as_json

    is64bit = '64-bit' if sys.maxsize > 2**32 else '32-bit'

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, args.show_ffmpeg_debug)

    if(args.debug and args.input == []):
        import platform

        print('Python Version: {} {}'.format(platform.python_version(), is64bit))
        print('Platform: {} {} {}'.format(platform.system(), platform.release(), platform.machine().lower()))
        print('FFmpeg path: {}'.format(ffmpeg.path))
        print('FFmpeg version: {}'.format(ffmpeg.version))
        print('Auto-Editor version {}'.format(auto_editor.version))
        sys.exit()

    if(is64bit == '32-bit'):
        Log().warning('You have the 32-bit version of Python, which may lead to '
            'memory crashes.')

    if(args.version):
        print('Auto-Editor version {}'.format(auto_editor.version))
        sys.exit()

    if(args.temp_dir is None):
        TEMP = tempfile.mkdtemp()
    else:
        TEMP = args.temp_dir
        if(os.path.isfile(TEMP)):
            Log().error('Temp directory cannot be an already existing file.')
        if(os.path.isdir(TEMP)):
            if(len(os.listdir(TEMP)) != 0):
                Log().error('Temp directory should be empty!')
        else:
            os.mkdir(TEMP)

    log = Log(args.debug, args.quiet, temp=TEMP)
    log.debug('Temp Directory: {}'.format(TEMP))

    if(args.input == []):
        log.error('You need to give auto-editor an input file or folder so it can '
            'do the work for you.')

    if([args.export_to_premiere, args.export_to_final_cut_pro, args.export_as_audio,
        args.export_to_shotcut, args.export_as_clip_sequence].count(True) > 1):
        log.error('You must choose only one export option.')

    if(isinstance(args.frame_margin, str)):
        try:
            if(float(args.frame_margin) < 0):
                log.error('Frame margin cannot be negative.')
        except ValueError:
            log.error('Frame margin {}, is not valid.'.format(args.frame_margin))
    elif(args.frame_margin < 0):
        log.error('Frame margin cannot be negative.')
    if(args.constant_rate_factor != 'unset'):
        if(int(args.constant_rate_factor) < 0 or int(args.constant_rate_factor) > 51):
            log.error('Constant rate factor (crf) must be between 0-51.')
    if(args.md_width < 1):
        log.error('--md_width cannot be less than 1.')
    if(args.md_dilates < 0):
        log.error('--md_dilates cannot be less than 0')

    def write_starting_message(args):
        if(args.export_to_premiere):
            return 'Exporting to Adobe Premiere Pro XML file.'
        if(args.export_to_final_cut_pro):
            return 'Exporting to Final Cut Pro XML file.'
        if(args.export_to_shotcut):
            return 'Exporting to Shotcut XML Timeline file.'
        if(args.export_as_audio):
            return 'Exporting as audio.'
        return 'Starting.'

    if(not args.preview):
        log.conwrite(write_starting_message(args))

    if(args.preview or args.export_as_clip_sequence or making_data_file):
        args.no_open = True

    if(args.md_blur < 0):
        args.md_blur = 0

    if(args.silent_speed <= 0 or args.silent_speed > 99999):
        args.silent_speed = 99999

    if(args.video_speed <= 0 or args.video_speed > 99999):
        args.video_speed = 99999

    if(args.output_file is None):
        args.output_file = []

    from auto_editor.validate_input import valid_input
    input_list, segments = valid_input(args.input, ffmpeg, args, log)

    if(len(args.output_file) < len(input_list)):
        for i in range(len(input_list) - len(args.output_file)):
            args.output_file.append(set_output_name(input_list[i], None,
                making_data_file, args))

    if(args.combine_files):
        if(exporting_to_editor):
            temp_file = 'combined.mp4'
        else:
            temp_file = os.path.join(TEMP, 'combined.mp4')

        cmd = []
        for fileref in input_list:
            cmd.extend(['-i', fileref])
        cmd.extend(['-filter_complex', '[0:v]concat=n={}:v=1:a=1'.format(len(input_list)),
            '-codec:v', 'h264', '-pix_fmt', 'yuv420p', '-strict', '-2', temp_file])
        ffmpeg.run(cmd)
        del cmd
        input_list = [temp_file]

    speeds = [args.silent_speed, args.video_speed]
    if(args.cut_out != [] and 99999 not in speeds):
        speeds.append(99999)

    for item in args.set_speed_for_range:
        if(item[0] not in speeds):
            speeds.append(float(item[0]))

    log.debug('Speeds: {}'.format(speeds))

    def main_loop(input_list, ffmpeg, args, speeds, segments, log):
        num_cuts = 0

        progress = ProgressBar(args.machine_readable_progress, args.no_progress)

        for i, input_path in enumerate(input_list):
            inp = ffmpeg.file_info(input_path)

            if(len(input_list) > 1):
                log.conwrite('Working on {}'.format(inp.basename))

            cuts, output_path = edit_media(i, inp, ffmpeg, args, progress, speeds,
                segments[i], exporting_to_editor, making_data_file, TEMP, log)
            num_cuts += cuts

        if(not args.preview and not making_data_file):
            timer.stop()

        if(not args.preview and making_data_file):
            # Assume making each cut takes about 30 seconds.
            time_save = usefulfunctions.human_readable_time(num_cuts * 30)
            s = 's' if num_cuts != 1 else ''

            log.print('Auto-Editor made {} cut{}, which would have taken about {} if '
                'edited manually.'.format(num_cuts, s, time_save))

        if(not args.no_open):
            usefulfunctions.open_with_system_default(output_path, log)

    try:
        main_loop(input_list, ffmpeg, args, speeds, segments, log)
    except KeyboardInterrupt:
        log.error('Keyboard Interrupt')
    log.cleanup()

if(__name__ == '__main__'):
    main()



'''subcommands/subdump.py'''

import sys

def subdump_options(parser):
    parser.add_argument('--ffmpeg_location', default=None,
        help='Point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='Use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='Print info about the program or an option and exit.')
    parser.add_argument('input', nargs='*',
        help='The path to a file you want inspected.')
    return parser

def main(sys_args=sys.argv[1:]):
    import os
    import tempfile

    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.log import Log
    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('subdump', auto_editor.version,
        description='Dump subtitle streams to stdout in text readable form.')
    parser = subdump_options(parser)

    temp = tempfile.mkdtemp()
    log = Log(temp=temp)

    try:
        args = parser.parse_args(sys_args)
    except vanparse.ParserError as e:
        log.error(str(e))

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, debug=False)

    for i, input_file in enumerate(args.input):
        inp = ffmpeg.file_info(input_file)

        cmd = ['-i', input_file]
        for s, sub in enumerate(inp.subtitle_streams):
            cmd.extend(['-map', '0:s:{}'.format(s),
                os.path.join(temp, '{}s{}.{}'.format(i, s, sub['ext']))])
        ffmpeg.run(cmd)

        for s, sub in enumerate(inp.subtitle_streams):
            print('file: {} ({}:{}:{})'.format(input_file, s, sub['lang'], sub['ext']))
            with open(os.path.join(temp, '{}s{}.{}'.format(i, s, sub['ext']))) as file:
                print(file.read())
            print('------')

    log.cleanup()

if(__name__ == '__main__'):
    main()



'''scipy/wavfile.py'''

# Mirrored from: github.com/scipy/scipy/blob/master/scipy/io/wavfile.py

import io
import numpy
import struct
from enum import IntEnum

class WAVE_FORMAT(IntEnum):
    UNKNOWN = 0x0000
    PCM = 0x0001
    ADPCM = 0x0002
    IEEE_FLOAT = 0x0003
    VSELP = 0x0004
    IBM_CVSD = 0x0005
    ALAW = 0x0006
    MULAW = 0x0007
    DTS = 0x0008
    DRM = 0x0009
    WMAVOICE9 = 0x000A
    WMAVOICE10 = 0x000B
    OKI_ADPCM = 0x0010
    DVI_ADPCM = 0x0011
    MEDIASPACE_ADPCM = 0x0012
    SIERRA_ADPCM = 0x0013
    G723_ADPCM = 0x0014
    DIGISTD = 0x0015
    DIGIFIX = 0x0016
    DIALOGIC_OKI_ADPCM = 0x0017
    MEDIAVISION_ADPCM = 0x0018
    CU_CODEC = 0x0019
    HP_DYN_VOICE = 0x001A
    YAMAHA_ADPCM = 0x0020
    SONARC = 0x0021
    DSPGROUP_TRUESPEECH = 0x0022
    ECHOSC1 = 0x0023
    AUDIOFILE_AF36 = 0x0024
    APTX = 0x0025
    AUDIOFILE_AF10 = 0x0026
    PROSODY_1612 = 0x0027
    LRC = 0x0028
    DOLBY_AC2 = 0x0030
    GSM610 = 0x0031
    MSNAUDIO = 0x0032
    ANTEX_ADPCME = 0x0033
    CONTROL_RES_VQLPC = 0x0034
    DIGIREAL = 0x0035
    DIGIADPCM = 0x0036
    CONTROL_RES_CR10 = 0x0037
    NMS_VBXADPCM = 0x0038
    CS_IMAADPCM = 0x0039
    ECHOSC3 = 0x003A
    ROCKWELL_ADPCM = 0x003B
    ROCKWELL_DIGITALK = 0x003C
    XEBEC = 0x003D
    G721_ADPCM = 0x0040
    G728_CELP = 0x0041
    MSG723 = 0x0042
    INTEL_G723_1 = 0x0043
    INTEL_G729 = 0x0044
    SHARP_G726 = 0x0045
    MPEG = 0x0050
    RT24 = 0x0052
    PAC = 0x0053
    MPEGLAYER3 = 0x0055
    LUCENT_G723 = 0x0059
    CIRRUS = 0x0060
    ESPCM = 0x0061
    VOXWARE = 0x0062
    CANOPUS_ATRAC = 0x0063
    G726_ADPCM = 0x0064
    G722_ADPCM = 0x0065
    DSAT = 0x0066
    DSAT_DISPLAY = 0x0067
    VOXWARE_BYTE_ALIGNED = 0x0069
    VOXWARE_AC8 = 0x0070
    VOXWARE_AC10 = 0x0071
    VOXWARE_AC16 = 0x0072
    VOXWARE_AC20 = 0x0073
    VOXWARE_RT24 = 0x0074
    VOXWARE_RT29 = 0x0075
    VOXWARE_RT29HW = 0x0076
    VOXWARE_VR12 = 0x0077
    VOXWARE_VR18 = 0x0078
    VOXWARE_TQ40 = 0x0079
    VOXWARE_SC3 = 0x007A
    VOXWARE_SC3_1 = 0x007B
    SOFTSOUND = 0x0080
    VOXWARE_TQ60 = 0x0081
    MSRT24 = 0x0082
    G729A = 0x0083
    MVI_MVI2 = 0x0084
    DF_G726 = 0x0085
    DF_GSM610 = 0x0086
    ISIAUDIO = 0x0088
    ONLIVE = 0x0089
    MULTITUDE_FT_SX20 = 0x008A
    INFOCOM_ITS_G721_ADPCM = 0x008B
    CONVEDIA_G729 = 0x008C
    CONGRUENCY = 0x008D
    SBC24 = 0x0091
    DOLBY_AC3_SPDIF = 0x0092
    MEDIASONIC_G723 = 0x0093
    PROSODY_8KBPS = 0x0094
    ZYXEL_ADPCM = 0x0097
    PHILIPS_LPCBB = 0x0098
    PACKED = 0x0099
    MALDEN_PHONYTALK = 0x00A0
    RACAL_RECORDER_GSM = 0x00A1
    RACAL_RECORDER_G720_A = 0x00A2
    RACAL_RECORDER_G723_1 = 0x00A3
    RACAL_RECORDER_TETRA_ACELP = 0x00A4
    NEC_AAC = 0x00B0
    RAW_AAC1 = 0x00FF
    RHETOREX_ADPCM = 0x0100
    IRAT = 0x0101
    VIVO_G723 = 0x0111
    VIVO_SIREN = 0x0112
    PHILIPS_CELP = 0x0120
    PHILIPS_GRUNDIG = 0x0121
    DIGITAL_G723 = 0x0123
    SANYO_LD_ADPCM = 0x0125
    SIPROLAB_ACEPLNET = 0x0130
    SIPROLAB_ACELP4800 = 0x0131
    SIPROLAB_ACELP8V3 = 0x0132
    SIPROLAB_G729 = 0x0133
    SIPROLAB_G729A = 0x0134
    SIPROLAB_KELVIN = 0x0135
    VOICEAGE_AMR = 0x0136
    G726ADPCM = 0x0140
    DICTAPHONE_CELP68 = 0x0141
    DICTAPHONE_CELP54 = 0x0142
    QUALCOMM_PUREVOICE = 0x0150
    QUALCOMM_HALFRATE = 0x0151
    TUBGSM = 0x0155
    MSAUDIO1 = 0x0160
    WMAUDIO2 = 0x0161
    WMAUDIO3 = 0x0162
    WMAUDIO_LOSSLESS = 0x0163
    WMASPDIF = 0x0164
    UNISYS_NAP_ADPCM = 0x0170
    UNISYS_NAP_ULAW = 0x0171
    UNISYS_NAP_ALAW = 0x0172
    UNISYS_NAP_16K = 0x0173
    SYCOM_ACM_SYC008 = 0x0174
    SYCOM_ACM_SYC701_G726L = 0x0175
    SYCOM_ACM_SYC701_CELP54 = 0x0176
    SYCOM_ACM_SYC701_CELP68 = 0x0177
    KNOWLEDGE_ADVENTURE_ADPCM = 0x0178
    FRAUNHOFER_IIS_MPEG2_AAC = 0x0180
    DTS_DS = 0x0190
    CREATIVE_ADPCM = 0x0200
    CREATIVE_FASTSPEECH8 = 0x0202
    CREATIVE_FASTSPEECH10 = 0x0203
    UHER_ADPCM = 0x0210
    ULEAD_DV_AUDIO = 0x0215
    ULEAD_DV_AUDIO_1 = 0x0216
    QUARTERDECK = 0x0220
    ILINK_VC = 0x0230
    RAW_SPORT = 0x0240
    ESST_AC3 = 0x0241
    GENERIC_PASSTHRU = 0x0249
    IPI_HSX = 0x0250
    IPI_RPELP = 0x0251
    CS2 = 0x0260
    SONY_SCX = 0x0270
    SONY_SCY = 0x0271
    SONY_ATRAC3 = 0x0272
    SONY_SPC = 0x0273
    TELUM_AUDIO = 0x0280
    TELUM_IA_AUDIO = 0x0281
    NORCOM_VOICE_SYSTEMS_ADPCM = 0x0285
    FM_TOWNS_SND = 0x0300
    MICRONAS = 0x0350
    MICRONAS_CELP833 = 0x0351
    BTV_DIGITAL = 0x0400
    INTEL_MUSIC_CODER = 0x0401
    INDEO_AUDIO = 0x0402
    QDESIGN_MUSIC = 0x0450
    ON2_VP7_AUDIO = 0x0500
    ON2_VP6_AUDIO = 0x0501
    VME_VMPCM = 0x0680
    TPC = 0x0681
    LIGHTWAVE_LOSSLESS = 0x08AE
    OLIGSM = 0x1000
    OLIADPCM = 0x1001
    OLICELP = 0x1002
    OLISBC = 0x1003
    OLIOPR = 0x1004
    LH_CODEC = 0x1100
    LH_CODEC_CELP = 0x1101
    LH_CODEC_SBC8 = 0x1102
    LH_CODEC_SBC12 = 0x1103
    LH_CODEC_SBC16 = 0x1104
    NORRIS = 0x1400
    ISIAUDIO_2 = 0x1401
    SOUNDSPACE_MUSICOMPRESS = 0x1500
    MPEG_ADTS_AAC = 0x1600
    MPEG_RAW_AAC = 0x1601
    MPEG_LOAS = 0x1602
    NOKIA_MPEG_ADTS_AAC = 0x1608
    NOKIA_MPEG_RAW_AAC = 0x1609
    VODAFONE_MPEG_ADTS_AAC = 0x160A
    VODAFONE_MPEG_RAW_AAC = 0x160B
    MPEG_HEAAC = 0x1610
    VOXWARE_RT24_SPEECH = 0x181C
    SONICFOUNDRY_LOSSLESS = 0x1971
    INNINGS_TELECOM_ADPCM = 0x1979
    LUCENT_SX8300P = 0x1C07
    LUCENT_SX5363S = 0x1C0C
    CUSEEME = 0x1F03
    NTCSOFT_ALF2CM_ACM = 0x1FC4
    DVM = 0x2000
    DTS2 = 0x2001
    MAKEAVIS = 0x3313
    DIVIO_MPEG4_AAC = 0x4143
    NOKIA_ADAPTIVE_MULTIRATE = 0x4201
    DIVIO_G726 = 0x4243
    LEAD_SPEECH = 0x434C
    LEAD_VORBIS = 0x564C
    WAVPACK_AUDIO = 0x5756
    OGG_VORBIS_MODE_1 = 0x674F
    OGG_VORBIS_MODE_2 = 0x6750
    OGG_VORBIS_MODE_3 = 0x6751
    OGG_VORBIS_MODE_1_PLUS = 0x676F
    OGG_VORBIS_MODE_2_PLUS = 0x6770
    OGG_VORBIS_MODE_3_PLUS = 0x6771
    ALAC = 0x6C61
    _3COM_NBX = 0x7000  # Can't have leading digit
    OPUS = 0x704F
    FAAD_AAC = 0x706D
    AMR_NB = 0x7361
    AMR_WB = 0x7362
    AMR_WP = 0x7363
    GSM_AMR_CBR = 0x7A21
    GSM_AMR_VBR_SID = 0x7A22
    COMVERSE_INFOSYS_G723_1 = 0xA100
    COMVERSE_INFOSYS_AVQSBC = 0xA101
    COMVERSE_INFOSYS_SBC = 0xA102
    SYMBOL_G729_A = 0xA103
    VOICEAGE_AMR_WB = 0xA104
    INGENIENT_G726 = 0xA105
    MPEG4_AAC = 0xA106
    ENCORE_G726 = 0xA107
    ZOLL_ASAO = 0xA108
    SPEEX_VOICE = 0xA109
    VIANIX_MASC = 0xA10A
    WM9_SPECTRUM_ANALYZER = 0xA10B
    WMF_SPECTRUM_ANAYZER = 0xA10C
    GSM_610 = 0xA10D
    GSM_620 = 0xA10E
    GSM_660 = 0xA10F
    GSM_690 = 0xA110
    GSM_ADAPTIVE_MULTIRATE_WB = 0xA111
    POLYCOM_G722 = 0xA112
    POLYCOM_G728 = 0xA113
    POLYCOM_G729_A = 0xA114
    POLYCOM_SIREN = 0xA115
    GLOBAL_IP_ILBC = 0xA116
    RADIOTIME_TIME_SHIFT_RADIO = 0xA117
    NICE_ACA = 0xA118
    NICE_ADPCM = 0xA119
    VOCORD_G721 = 0xA11A
    VOCORD_G726 = 0xA11B
    VOCORD_G722_1 = 0xA11C
    VOCORD_G728 = 0xA11D
    VOCORD_G729 = 0xA11E
    VOCORD_G729_A = 0xA11F
    VOCORD_G723_1 = 0xA120
    VOCORD_LBC = 0xA121
    NICE_G728 = 0xA122
    FRACE_TELECOM_G729 = 0xA123
    CODIAN = 0xA124
    FLAC = 0xF1AC
    EXTENSIBLE = 0xFFFE
    DEVELOPMENT = 0xFFFF


KNOWN_WAVE_FORMATS = {WAVE_FORMAT.PCM, WAVE_FORMAT.IEEE_FLOAT}


def _raise_bad_format(format_tag):
    try:
        format_name = WAVE_FORMAT(format_tag).name
    except ValueError:
        format_name = '{:#06x}'.format(format_tag)
    raise ValueError("Unknown wave file format: {}. Supported ".format(format_name) +
        "formats: " + ', '.join(x.name for x in KNOWN_WAVE_FORMATS))


def _read_fmt_chunk(fid, is_big_endian):
    if is_big_endian:
        fmt = '>'
    else:
        fmt = '<'

    size = struct.unpack(fmt+'I', fid.read(4))[0]

    if size < 16:
        raise ValueError("Binary structure of wave file is not compliant")

    res = struct.unpack(fmt+'HHIIHH', fid.read(16))
    bytes_read = 16

    format_tag, channels, fs, bytes_per_second, block_align, bit_depth = res

    if format_tag == WAVE_FORMAT.EXTENSIBLE and size >= (16+2):
        ext_chunk_size = struct.unpack(fmt+'H', fid.read(2))[0]
        bytes_read += 2
        if ext_chunk_size >= 22:
            extensible_chunk_data = fid.read(22)
            bytes_read += 22
            raw_guid = extensible_chunk_data[2+4:2+4+16]
            # GUID template {XXXXXXXX-0000-0010-8000-00AA00389B71} (RFC-2361)
            # MS GUID byte order: first three groups are native byte order,
            # rest is Big Endian
            if is_big_endian:
                tail = b'\x00\x00\x00\x10\x80\x00\x00\xAA\x00\x38\x9B\x71'
            else:
                tail = b'\x00\x00\x10\x00\x80\x00\x00\xAA\x00\x38\x9B\x71'
            if raw_guid.endswith(tail):
                format_tag = struct.unpack(fmt+'I', raw_guid[:4])[0]
        else:
            raise ValueError("Binary structure of wave file is not compliant")

    if format_tag not in KNOWN_WAVE_FORMATS:
        _raise_bad_format(format_tag)

    # move file pointer to next chunk
    if size > bytes_read:
        fid.read(size - bytes_read)

    # fmt should always be 16, 18 or 40, but handle it just in case
    _handle_pad_byte(fid, size)

    return (size, format_tag, channels, fs, bytes_per_second, block_align,
            bit_depth)


def _read_data_chunk(fid, format_tag, channels, bit_depth, is_big_endian,
                     block_align, mmap=False):
    if is_big_endian:
        fmt = '>'
    else:
        fmt = '<'

    # Size of the data subchunk in bytes
    size = struct.unpack(fmt+'I', fid.read(4))[0]

    # Number of bytes per sample (sample container size)
    bytes_per_sample = block_align // channels
    n_samples = size // bytes_per_sample

    if format_tag == WAVE_FORMAT.PCM:
        if 1 <= bit_depth <= 8:
            dtype = 'u1'  # WAV of 8-bit integer or less are unsigned
        elif bytes_per_sample in {3, 5, 6, 7}:
            # No compatible dtype.  Load as raw bytes for reshaping later.
            dtype = 'V1'
        elif bit_depth <= 64:
            # Remaining bit depths can map directly to signed numpy dtypes
            dtype = '{}i{}'.format(fmt, bytes_per_sample)
        else:
            raise ValueError("Unsupported bit depth: the WAV file "
                "has {}-bit integer data.".format(bit_depth))
    elif format_tag == WAVE_FORMAT.IEEE_FLOAT:
        if bit_depth in {32, 64}:
            dtype = '{}f{}'.format(fmt, bytes_per_sample)
        else:
            raise ValueError("Unsupported bit depth: the WAV file "
                "has {}-bit floating-point data.".format(bit_depth))
    else:
        _raise_bad_format(format_tag)

    start = fid.tell()
    if not mmap:
        try:
            count = size if dtype == 'V1' else n_samples
            data = numpy.fromfile(fid, dtype=dtype, count=count)
        except io.UnsupportedOperation:  # not a C-like file
            fid.seek(start, 0)  # just in case it seeked, though it shouldn't
            data = numpy.frombuffer(fid.read(size), dtype=dtype)

        if dtype == 'V1':
            # Rearrange raw bytes into smallest compatible numpy dtype
            dt = numpy.int32 if bytes_per_sample == 3 else numpy.int64
            a = numpy.zeros((len(data) // bytes_per_sample, dt().itemsize),
                            dtype='V1')
            a[:, -bytes_per_sample:] = data.reshape((-1, bytes_per_sample))
            data = a.view(dt).reshape(a.shape[:-1])
    else:
        if bytes_per_sample in {1, 2, 4, 8}:
            start = fid.tell()
            data = numpy.memmap(fid, dtype=dtype, mode='c', offset=start,
                                shape=(n_samples,))
            fid.seek(start + size)
        else:
            raise ValueError("mmap=True not compatible with "
                "{}-byte container size.".format(bytes_per_sample))

    _handle_pad_byte(fid, size)

    if channels > 1:
        data = data.reshape(-1, channels)
    return data


def _skip_unknown_chunk(fid, is_big_endian):
    if is_big_endian:
        fmt = '>I'
    else:
        fmt = '<I'

    data = fid.read(4)
    # call unpack() and seek() only if we have really read data from file
    # otherwise empty read at the end of the file would trigger
    # unnecessary exception at unpack() call
    # in case data equals somehow to 0, there is no need for seek() anyway
    if data:
        size = struct.unpack(fmt, data)[0]
        fid.seek(size, 1)
        _handle_pad_byte(fid, size)


def _read_riff_chunk(fid):
    str1 = fid.read(4)  # File signature
    if str1 == b'RIFF':
        is_big_endian = False
        fmt = '<I'
    elif str1 == b'RIFX':
        is_big_endian = True
        fmt = '>I'
    else:
        # There are also .wav files with "FFIR" or "XFIR" signatures?
        raise ValueError("File format {} not understood. ".format(repr(str1)) +
            "Only 'RIFF' and 'RIFX' supported.")

    # Size of entire file
    file_size = struct.unpack(fmt, fid.read(4))[0] + 8

    str2 = fid.read(4)
    if str2 != b'WAVE':
        raise ValueError("Not a WAV file. RIFF form type is {}.".format(repr(str2)))

    return file_size, is_big_endian


def _handle_pad_byte(fid, size):
    # "If the chunk size is an odd number of bytes, a pad byte with value zero
    # is written after ckData." So we need to seek past this after each chunk.
    if size % 2:
        fid.seek(1, 1)


def read(filename, mmap=False):
    if hasattr(filename, 'read'):
        fid = filename
        mmap = False
    else:
        fid = open(filename, 'rb')

    try:
        file_size, is_big_endian = _read_riff_chunk(fid)
        fmt_chunk_received = False
        data_chunk_received = False
        while fid.tell() < file_size:
            # read the next chunk
            chunk_id = fid.read(4)

            if not chunk_id:
                if data_chunk_received:
                    # End of file but data successfully read
                    break
                else:
                    raise ValueError("Unexpected end of file.")
            elif len(chunk_id) < 4:
                msg = "Incomplete chunk ID: {}".format(repr(chunk_id))
                # If we have the data, ignore the broken chunk
                if fmt_chunk_received and data_chunk_received:
                    pass
                else:
                    raise ValueError(msg)

            if chunk_id == b'fmt ':
                fmt_chunk_received = True
                fmt_chunk = _read_fmt_chunk(fid, is_big_endian)
                format_tag, channels, fs = fmt_chunk[1:4]
                bit_depth = fmt_chunk[6]
                block_align = fmt_chunk[5]
            elif chunk_id == b'fact':
                _skip_unknown_chunk(fid, is_big_endian)
            elif chunk_id == b'data':
                data_chunk_received = True
                if not fmt_chunk_received:
                    raise ValueError("No fmt chunk before data")
                data = _read_data_chunk(fid, format_tag, channels, bit_depth,
                                        is_big_endian, block_align, mmap)
            elif chunk_id == b'LIST':
                # Someday this could be handled properly but for now skip it
                _skip_unknown_chunk(fid, is_big_endian)
            elif chunk_id in {b'JUNK', b'Fake'}:
                # Skip alignment chunks without warning
                _skip_unknown_chunk(fid, is_big_endian)
            else:
                # Chunk (non-data) not understood, skipping it.",
                _skip_unknown_chunk(fid, is_big_endian)
    finally:
        if not hasattr(filename, 'read'):
            fid.close()
        else:
            fid.seek(0)

    return fs, data



'''audiotsm2/base/tsm.py'''

class TSM():
    def run(self, reader, writer, flush=True):
        finished = False
        while not (finished and reader.empty):
            self.read_from(reader)
            _, finished = self.write_to(writer)

        if flush:
            finished = False
            while not finished:
                _, finished = self.flush_to(writer)

            self.clear()



'''subcommands/info.py'''

import sys
import os.path

def info_options(parser):
    parser.add_argument('--include_vfr', '--has_vfr', action='store_true',
        help='skip information that is very slow to get.')
    parser.add_argument('--ffmpeg_location', default=None,
        help='point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    parser.add_argument('input', nargs='*',
        help='the path to a file you want inspected.')
    return parser

def main(sys_args=sys.argv[1:]):

    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.func import aspect_ratio
    from auto_editor.utils.log import Log

    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('info', auto_editor.version,
        description='Get basic information about media files.')
    parser = info_options(parser)

    log = Log()

    try:
        args = parser.parse_args(sys_args)
    except vanparse.ParserError as e:
        log.error(str(e))

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, False)

    def aspect_str(w, h):
        w, h = aspect_ratio(int(w), int(h))
        if(w is None):
            return ''
        return ' ({}:{})'.format(w, h)

    for file in args.input:
        text = ''
        if(os.path.exists(file)):
            text += 'file: {}\n'.format(file)
        else:
            log.error('Could not find file: {}'.format(file))

        inp = ffmpeg.file_info(file)

        if(len(inp.video_streams) > 0):
            text += f' - video tracks: {len(inp.video_streams)}\n'

        for track, stream in enumerate(inp.video_streams):
            text += '   - Track #{}\n'.format(track)

            text += '     - codec: {}\n'.format(stream['codec'])
            if(stream['fps'] is not None):
                text += '     - fps: {}\n'.format(stream['fps'])

            w = stream['width']
            h = stream['height']

            if(w is not None and h is not None):
                text += '     - resolution: {}x{}{}\n'.format(w, h, aspect_str(w, h))

            if(stream['bitrate'] is not None):
                text += '     - bitrate: {}\n'.format(stream['bitrate'])
            if(stream['lang'] is not None):
                text += '     - lang: {}\n'.format(stream['lang'])


        if(len(inp.audio_streams) > 0):
            text += f' - audio tracks: {len(inp.audio_streams)}\n'

        for track, stream in enumerate(inp.audio_streams):
            text += '   - Track #{}\n'.format(track)
            text += '     - codec: {}\n'.format(stream['codec'])
            text += '     - samplerate: {}\n'.format(stream['samplerate'])

            if(stream['bitrate'] is not None):
                text += '     - bitrate: {}\n'.format(stream['bitrate'])

            if(stream['lang'] is not None):
                text += '     - lang: {}\n'.format(stream['lang'])

        if(len(inp.subtitle_streams) > 0):
            text += f' - subtitle tracks: {len(inp.subtitle_streams)}\n'

        for track, stream in enumerate(inp.subtitle_streams):
            text += '   - Track #{}\n'.format(track)
            text += '     - codec: {}\n'.format(stream['codec'])
            if(stream['lang'] is not None):
                text += '     - lang: {}\n'.format(stream['lang'])


        if(len(inp.video_streams) == 0 and len(inp.audio_streams) == 0 and
            len(inp.subtitle_streams) == 0):
            text += 'Invalid media.\n'
        else:
            text += ' - container:\n'
            if(inp.duration is not None):
                text += '   - duration: {}\n'.format(inp.duration)
            if(inp.bitrate is not None):
                text += '   - bitrate: {}\n'.format(inp.bitrate)

            if(args.include_vfr):
                print(text, end='')
                text = ''
                fps_mode = ffmpeg.pipe(['-i', file, '-hide_banner', '-vf', 'vfrdet',
                    '-an', '-f', 'null', '-'])
                fps_mode = fps_mode.strip()

                if('VFR:' in fps_mode):
                    fps_mode = (fps_mode[fps_mode.index('VFR:'):]).strip()

                text += '  - {}\n'.format(fps_mode)

        print(text)

if(__name__ == '__main__'):
    main()



'''render/subtitle.py'''

import os
import re

class SubtitleParser:
    def __init__(self):
        self.supported_codecs = ['ass', 'webvtt', 'mov_text']

    def parse(self, text, fps, codec):

        if(codec not in self.supported_codecs):
            raise ValueError('codec {} not supported.'.format(codec))

        self.fps = fps
        self.codec = codec
        self.contents = []

        if(codec == 'ass'):
            time_code = re.compile(r'(.*)(\d+:\d+:[\d.]+)(.*)(\d+:\d+:[\d.]+)(.*)')
        if(codec == 'webvtt'):
            time_code = re.compile(r'()(\d+:[\d.]+)( --> )(\d+:[\d.]+)(\n.*)')
        if(codec == 'mov_text'):
            time_code = re.compile(r'()(\d+:\d+:[\d,]+)( --> )(\d+:\d+:[\d,]+)(\n.*)')

        for i, item in enumerate(re.finditer(time_code, text)):
            if(i == 0):
                self.header = text[:item.span()[0]]

            self.contents.append(
                [self.to_frame(item.group(2)), self.to_frame(item.group(4)),
                item.group(1), item.group(3), item.group(5) + '\n']
            )

        self.footer = text[item.span()[1]:]


    def edit(self, chunks, speeds):

        for cut in reversed(chunks):
            the_speed = speeds[cut[2]]
            speed_factor = 1 if the_speed == 99999 else 1 - (1 / the_speed)

            new_content = []
            for content in self.contents:
                if(cut[0] <= content[1] and cut[1] > content[0]):

                    diff = int(
                        (min(cut[1], content[1]) - max(cut[0], content[0])) * speed_factor
                    )
                    if(content[0] > cut[0]):
                        content[0] -= diff
                        content[1] -= diff

                    content[1] -= diff

                elif(content[0] >= cut[0]):
                    diff = int((cut[1] - cut[0]) * speed_factor)

                    content[0] -= diff
                    content[1] -= diff

                if(content[0] != content[1]):
                    new_content.append(content)

        self.contents = new_content

    def write(self, file_path):
        with open(file_path, 'w') as file:
            file.write(self.header)
            for item in self.contents:
                file.write('{before}{start_time}{middle}{end_time}{after}'.format(
                    before=item[2],
                    start_time=self.to_timecode(item[0]),
                    middle=item[3],
                    end_time=self.to_timecode(item[1]),
                    after=item[4],
                ))
            file.write(self.footer)

    def to_frame(self, text):
        # type: (str) -> int
        if(self.codec == 'mov_text'):
            time_format = r'(\d+):?(\d+):([\d,]+)'
        else:
            time_format = r'(\d+):?(\d+):([\d.]+)'

        nums = re.match(time_format, text)
        hours, minutes, seconds = nums.groups()
        seconds = seconds.replace(',', '.', 1)
        return round((int(hours) * 3600 + int(minutes) * 60 + float(seconds)) * self.fps)

    def to_timecode(self, frame):
        # type: (int) -> str
        seconds = frame / self.fps

        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)

        sig_fig = 2 if self.codec == 'ass' else 3
        s = str(round(s, sig_fig)).zfill(2)

        if(self.codec == 'webvtt'):
            if(int(h) == 0):
                return '{:02d}:{}'.format(int(m), s)
            time_format = '{:02d}:{:02d}:{}'
        elif(self.codec == 'mov_text'):
            s = s.replace('.', ',', 1)
            time_format = '{:02d}:{:02d}:{}'
        else:
            time_format = '{:d}:{:02d}:{}'

        return time_format.format(int(h), int(m), s)


def cut_subtitles(ffmpeg, inp, chunks, speeds, fps, temp, log):
    for s, sub in enumerate(inp.subtitle_streams):
        file_path = os.path.join(temp, '{}s.{}'.format(s, sub['ext']))
        new_path = os.path.join(temp, 'new{}s.{}'.format(s, sub['ext']))

        parser = SubtitleParser()

        if(sub['codec'] in parser.supported_codecs):
            with open(file_path, 'r') as file:
                parser.parse(file.read(), fps, sub['codec'])
        else:
            convert_path = os.path.join(temp, '{}s_convert.vtt'.format(s))
            ffmpeg.run(['-i', file_path, convert_path])
            with open(convert_path, 'r') as file:
                parser.parse(file.read(), fps, 'webvtt')

        parser.edit(chunks, speeds)
        parser.write(new_path)



'''analyze/motion.py'''

# Motion detection method is based on this post:
# pyimagesearch.com/2015/05/25/basic-motion-detection-and-tracking-with-python-and-opencv/

import cv2
import numpy as np

def resize(image, width=None, height=None, inter=cv2.INTER_AREA):
    if(width is None and height is None):
        return image

    h, w = image.shape[:2]
    if(width is None):
        r = height / h
        dim = (int(w * r), height)
    else:
        r = width / w
        dim = (width, int(h * r))

    return cv2.resize(image, dim, interpolation=inter)


def display_motion_levels(inp, width, dilates, blur):
    import sys

    cap = cv2.VideoCapture(inp.path)

    prev_frame = None
    gray = None
    total_pixels = None

    while cap.isOpened():
        if(gray is None):
            prev_frame = None
        else:
            prev_frame = gray

        ret, frame = cap.read()
        if(not ret):
            break

        frame = resize(frame, width=width)
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        if(blur > 0):
            gray = cv2.GaussianBlur(gray, (blur, blur), 0)

        if(prev_frame is not None):
            frame_delta = cv2.absdiff(prev_frame, gray)
            thresh = cv2.threshold(frame_delta, 25, 255, cv2.THRESH_BINARY)[1]

            if(dilates > 0):
                thresh = cv2.dilate(thresh, None, iterations=dilates)

            if(total_pixels is None):
                total_pixels = thresh.shape[0] * thresh.shape[1]

            sys.stdout.write('{:.20f}\n'.format(np.count_nonzero(thresh) / total_pixels))

    cap.release()
    cv2.destroyAllWindows()


def motion_detection(inp, threshold, log, width, dilates, blur):
    # type: (Any, float, Any, int, int, int) -> np.ndarray

    from auto_editor.utils.progressbar import ProgressBar

    log.conwrite('Analyzing video motion.')

    cap = cv2.VideoCapture(inp.path)

    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) + 1

    log.debug('   - Cutting total_frames: {}'.format(total_frames))
    prev_frame = None
    gray = None
    has_motion = np.zeros((total_frames), dtype=np.bool_)
    total_pixels = None

    progress = ProgressBar(total_frames, 'Detecting motion')

    while cap.isOpened():
        if(gray is None):
            prev_frame = None
        else:
            prev_frame = gray

        ret, frame = cap.read()

        if(not ret):
            break

        cframe = int(cap.get(cv2.CAP_PROP_POS_FRAMES)) # current frame

        frame = resize(frame, width=width)
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY) # Convert frame to grayscale.
        if(blur > 0):
            gray = cv2.GaussianBlur(gray, (blur, blur), 0)

        if(prev_frame is not None):
            frame_delta = cv2.absdiff(prev_frame, gray)
            thresh = cv2.threshold(frame_delta, 25, 255, cv2.THRESH_BINARY)[1]

            # Dilate the thresholded image to fill in holes.
            if(dilates > 0):
                thresh = cv2.dilate(thresh, None, iterations=dilates)

            if(total_pixels is None):
                total_pixels = thresh.shape[0] * thresh.shape[1]

            if(np.count_nonzero(thresh) / total_pixels >= threshold):
                has_motion[cframe] = True

        progress.tick(cframe)

    cap.release()
    cv2.destroyAllWindows()

    log.conwrite('')

    return has_motion



'''preview.py'''

# Internal Libraries
from datetime import timedelta

# Included Libraries
from auto_editor.utils.func import get_new_length

def display_length(secs):
    # display length
    if(secs < 0):
        return '-' + str(timedelta(seconds=round(abs(secs))))
    return str(timedelta(seconds=round(secs)))

def time_frame(title, frames, fps):
    in_sec = round(frames / fps, 1)
    minutes = timedelta(seconds=round(in_sec))
    print('{}: {} secs ({})'.format(title, in_sec, minutes))


def preview(inp, chunks, speeds, log):
    fps = 30 if inp.fps is None else float(inp.fps)

    log.conwrite('')

    old_length = chunks[-1][1] / fps
    new_length = get_new_length(chunks, speeds, fps)

    diff = new_length - old_length

    print('\nlength:\n - change: ({}) 100% -> ({}) {}%\n - diff: ({}) {}%'.format(
        display_length(old_length),
        display_length(new_length),
        round((new_length / old_length) * 100, 2),
        display_length(diff),
        round((diff / old_length) * 100, 2),
    ))

    clips = 0
    cuts = 0
    cut_lens = []
    clip_lens = []
    for chunk in chunks:
        state = chunk[2]
        if(speeds[state] != 99999):
            clips += 1
            leng = (chunk[1] - chunk[0]) / speeds[state]
            clip_lens.append(leng)
        else:
            cuts += 1
            leng = chunk[1] - chunk[0]
            cut_lens.append(leng)

    print('clips: {}'.format(clips))
    if(len(clip_lens) == 1):
        time_frame(' - clip length', clip_lens[0], fps)
    else:
        time_frame(' - smallest', min(clip_lens), fps)
        time_frame(' - largest', max(clip_lens), fps)
        time_frame(' - average', sum(clip_lens) / len(clip_lens), fps)
        print('cuts: {}'.format(cuts))

    if(cut_lens != []):
        if(len(cut_lens) == 1):
            time_frame(' - cut length', cut_lens[0], fps)
        else:
            time_frame(' - smallest', min(cut_lens), fps)
            time_frame(' - largest', max(cut_lens), fps)
            time_frame(' - average', sum(cut_lens) / len(cut_lens), fps)
        print('')

    log.debug('Chunks: {}'.format(chunks))



'''formats/premiere.py'''

import os
from shutil import move

from .utils import indent, get_width_height, safe_mkdir

pixelar = 'square' # pixel aspect ratio
colordepth = '24'
ntsc = 'FALSE'
ana = 'FALSE' # anamorphic
depth = '16'

def fix_url(path):
    # type: (str) -> str
    from urllib.parse import quote
    from platform import system
    from os.path import abspath

    if(system() == 'Windows'):
        return'file://localhost/' + quote(abspath(path)).replace('%5C', '/')
    return 'file://localhost' + abspath(path)

def speedup(speed):
    # type: (...) -> str
    return indent(6, '<filter>', '\t<effect>', '\t\t<name>Time Remap</name>',
        '\t\t<effectid>timeremap</effectid>',
        '\t\t<effectcategory>motion</effectcategory>',
        '\t\t<effecttype>motion</effecttype>',
        '\t\t<mediatype>video</mediatype>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>variablespeed</parameterid>',
        '\t\t\t<name>variablespeed</name>', '\t\t\t<valuemin>0</valuemin>',
        '\t\t\t<valuemax>1</valuemax>',
        '\t\t\t<value>0</value>',
        '\t\t</parameter>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>speed</parameterid>',  '\t\t\t<name>speed</name>',
        '\t\t\t<valuemin>-100000</valuemin>', '\t\t\t<valuemax>100000</valuemax>',
        '\t\t\t<value>{}</value>'.format(speed),
        '\t\t</parameter>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>reverse</parameterid>',
        '\t\t\t<name>reverse</name>', '\t\t\t<value>FALSE</value>',
        '\t\t</parameter>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>frameblending</parameterid>',
        '\t\t\t<name>frameblending</name>', '\t\t\t<value>FALSE</value>',
        '\t\t</parameter>', '\t</effect>', '</filter>')

def handle_video_clips(outfile, clips, inp, timebase, duration, width, height, sr,
    pathurls):
    tracks = len(inp.audio_streams)
    total = 0
    for j, clip in enumerate(clips):
        my_start = int(total)
        total += (clip[1] - clip[0]) / (clip[2] / 100)
        my_end = int(total)

        outfile.write(indent(5, '<clipitem id="clipitem-{}">'.format(j+1),
            '\t<masterclipid>masterclip-2</masterclipid>',
            '\t<name>{}</name>'.format(inp.basename),
            '\t<start>{}</start>'.format(my_start),
            '\t<end>{}</end>'.format(my_end),
            '\t<in>{}</in>'.format(int(clip[0] / (clip[2] / 100))),
            '\t<out>{}</out>'.format(int(clip[1] / (clip[2] / 100)))))

        if(j == 0):
            outfile.write(indent(6, '<file id="file-1">',
                '\t<name>{}</name>'.format(inp.basename),
                '\t<pathurl>{}</pathurl>'.format(pathurls[0]),
                '\t<rate>',
                '\t\t<timebase>{}</timebase>'.format(timebase),
                '\t\t<ntsc>{}</ntsc>'.format(ntsc),
                '\t</rate>',
                '\t<duration>{}</duration>'.format(duration),
                '\t<media>', '\t\t<video>',
                '\t\t\t<samplecharacteristics>',
                '\t\t\t\t<rate>',
                '\t\t\t\t\t<timebase>{}</timebase>'.format(timebase),
                '\t\t\t\t\t<ntsc>{}</ntsc>'.format(ntsc),
                '\t\t\t\t</rate>',
                '\t\t\t\t<width>{}</width>'.format(width),
                '\t\t\t\t<height>{}</height>'.format(height),
                '\t\t\t\t<anamorphic>{}</anamorphic>'.format(ana),
                '\t\t\t\t<pixelaspectratio>{}</pixelaspectratio>'.format(pixelar),
                '\t\t\t\t<fielddominance>none</fielddominance>',
                '\t\t\t</samplecharacteristics>',
                '\t\t</video>', '\t\t<audio>',
                '\t\t\t<samplecharacteristics>',
                '\t\t\t\t<depth>{}</depth>'.format(depth),
                '\t\t\t\t<samplerate>{}</samplerate>'.format(sr),
                '\t\t\t</samplecharacteristics>',
                '\t\t\t<channelcount>2</channelcount>',
                '\t\t</audio>', '\t</media>', '</file>'))
        else:
            outfile.write('\t\t\t\t\t\t<file id="file-1"/>\n')

        if(clip[2] != 100):
            outfile.write(speedup(clip[2]))

        # Linking for video blocks
        for i in range(max(3, tracks + 1)):
            outfile.write('\t\t\t\t\t\t<link>\n')
            outfile.write('\t\t\t\t\t\t\t<linkclipref>clipitem-{}</linkclipref>\n'.format((i*(len(clips)))+j+1))
            if(i == 0):
                outfile.write('\t\t\t\t\t\t\t<mediatype>video</mediatype>\n')
            else:
                outfile.write('\t\t\t\t\t\t\t<mediatype>audio</mediatype>\n')
            if(i == 2):
                outfile.write('\t\t\t\t\t\t\t<trackindex>2</trackindex>\n')
            else:
                outfile.write('\t\t\t\t\t\t\t<trackindex>1</trackindex>\n')
            outfile.write('\t\t\t\t\t\t\t<clipindex>{}</clipindex>\n'.format(j+1))
            if(i > 0):
                outfile.write('\t\t\t\t\t\t\t<groupindex>1</groupindex>\n')
            outfile.write('\t\t\t\t\t\t</link>\n')

        outfile.write('\t\t\t\t\t</clipitem>\n')
    outfile.write(indent(3, '\t</track>', '</video>'))

def handle_audio_clips(tracks, outfile, audioFile, clips, inp, timebase, sr, pathurls):
    for t in range(tracks):
        outfile.write('\t\t\t\t<track currentExplodedTrackIndex="0" premiereTrackType="Stereo">\n')
        total = 0
        for j, clip in enumerate(clips):

            my_start = int(total)
            total += (clip[1] - clip[0]) / (clip[2] / 100)
            my_end = int(total)

            if(audioFile):
                clip_item_num = j + 1
                master_id = '1'
            else:
                clip_item_num = len(clips) + 1 + j + (t * len(clips))
                master_id = '2'

            outfile.write(indent(5,
                '<clipitem id="clipitem-{}" premiereChannelType="stereo">'.format(clip_item_num),
                '\t<masterclipid>masterclip-{}</masterclipid>'.format(master_id),
                '\t<name>{}</name>'.format(inp.basename),
                '\t<start>{}</start>'.format(my_start),
                '\t<end>{}</end>'.format(my_end),
                '\t<in>{}</in>'.format(int(clip[0] / (clip[2] / 100))),
                '\t<out>{}</out>'.format(int(clip[1] / (clip[2] / 100)))))

            if((audioFile and j == 0) or (t > 0 and j == 0)):
                outfile.write(indent(6, '<file id="file-{}">'.format(t+1),
                    '\t<name>{}</name>'.format(inp.basename),
                    '\t<pathurl>{}</pathurl>'.format(pathurls[t]),
                    '\t<rate>',
                    '\t\t<timebase>{}</timebase>'.format(timebase),
                    '\t\t<ntsc>{}</ntsc>'.format(ntsc),
                    '\t</rate>',
                    '\t<media>',
                    '\t\t<audio>',
                    '\t\t\t<samplecharacteristics>',
                    '\t\t\t\t<depth>{}</depth>'.format(depth),
                    '\t\t\t\t<samplerate>{}</samplerate>'.format(sr),
                    '\t\t\t</samplecharacteristics>',
                    '\t\t\t<channelcount>2</channelcount>',
                    '\t\t</audio>', '\t</media>', '</file>'))
            else:
                outfile.write('\t\t\t\t\t\t<file id="file-{}"/>\n'.format(t+1))

            outfile.write(indent(6, '<sourcetrack>',
                '\t<mediatype>audio</mediatype>',
                '\t<trackindex>1</trackindex>',
                '</sourcetrack>'))

            if(audioFile):
                outfile.write('\t\t\t\t\t</clipitem>\n')
            else:
                outfile.write(indent(6, '<labels>', '\t<label2>Iris</label2>', '</labels>'))

            # Add speed effect for audio blocks
            if(clip[2] != 100):
                outfile.write(speedup(clip[2]))

            outfile.write('\t\t\t\t\t</clipitem>\n')
        if(not audioFile):
            outfile.write('\t\t\t\t\t<outputchannelindex>1</outputchannelindex>\n')
        outfile.write('\t\t\t\t</track>\n')

def premiere_xml(inp, temp, output, clips, chunks, sampleRate, audioFile, fps, log):

    duration = chunks[len(chunks) - 1][1]
    pathurls = [fix_url(inp.path)]

    tracks = len(inp.audio_streams)

    log.debug('tracks: {}'.format(tracks))
    log.debug(inp.dirname)

    if(tracks > 1):
        name_without_extension = inp.basename[:inp.basename.rfind('.')]

        fold = safe_mkdir(os.path.join(inp.dirname, '{}_tracks'.format(
            name_without_extension)))

        for i in range(1, tracks):
            newtrack = os.path.join(fold, '{}.wav'.format(i))
            move(os.path.join(temp, '{}.wav'.format(i)), newtrack)
            pathurls.append(fix_url(newtrack))

    width, height = get_width_height(inp)
    if(width is None or height is None):
        width, height = '1280', '720'

    timebase = str(int(fps))

    groupName = 'Auto-Editor {} Group'.format('Audio' if audioFile else 'Video')

    with open(output, 'w', encoding='utf-8') as outfile:
        outfile.write('<?xml version="1.0" encoding="UTF-8"?>\n<!DOCTYPE xmeml>\n')
        outfile.write('<xmeml version="4">\n')
        outfile.write('\t<sequence>\n')
        outfile.write('\t\t<name>{}</name>\n'.format(groupName))

        if(audioFile):
            outfile.write('\t\t<duration>{}</duration>\n'.format(duration))

        outfile.write('\t\t<rate>\n')
        outfile.write('\t\t\t<timebase>{}</timebase>\n'.format(timebase))
        outfile.write('\t\t\t<ntsc>{}</ntsc>\n'.format(ntsc))
        outfile.write('\t\t</rate>\n')
        outfile.write('\t\t<media>\n')

        if(audioFile):
            outfile.write(indent(3, '<video>', '\t<format>',
                '\t\t<samplecharacteristics>',
                '\t\t\t<width>{}</width>'.format(width),
                '\t\t\t<height>{}</height>'.format(height),
                '\t\t\t<pixelaspectratio>{}</pixelaspectratio>'.format(pixelar),
                '\t\t\t<rate>',
                '\t\t\t\t<timebase>{}</timebase>'.format(timebase),
                '\t\t\t\t<ntsc>{}</ntsc>'.format(ntsc),
                '\t\t\t</rate>',
                '\t\t</samplecharacteristics>',
                '\t</format>', '</video>'))
        else:
            outfile.write(indent(3, '<video>', '\t<format>',
                '\t\t<samplecharacteristics>',
                '\t\t\t<rate>',
                '\t\t\t\t<timebase>{}</timebase>'.format(timebase),
                '\t\t\t\t<ntsc>{}</ntsc>'.format(ntsc),
                '\t\t\t</rate>',
                '\t\t\t<width>{}</width>'.format(width),
                '\t\t\t<height>{}</height>'.format(height),
                '\t\t\t<anamorphic>{}</anamorphic>'.format(ana),
                '\t\t\t<pixelaspectratio>{}</pixelaspectratio>'.format(pixelar),
                '\t\t\t<fielddominance>none</fielddominance>',
                '\t\t\t<colordepth>{}</colordepth>'.format(colordepth),
                '\t\t</samplecharacteristics>',
                '\t</format>',
                '\t<track>'))
            handle_video_clips(outfile, clips, inp, timebase, duration, width, height,
                sampleRate, pathurls)

        # Audio Clips
        outfile.write(indent(3, '<audio>',
            '\t<numOutputChannels>2</numOutputChannels>',
            '\t<format>',
            '\t\t<samplecharacteristics>',
            '\t\t\t<depth>{}</depth>'.format(depth),
            '\t\t\t<samplerate>{}</samplerate>'.format(sampleRate),
            '\t\t</samplecharacteristics>',
            '\t</format>'))

        print('')
        handle_audio_clips(tracks, outfile, audioFile, clips, inp, timebase, sampleRate,
            pathurls)

        outfile.write('\t\t\t</audio>\n')
        outfile.write('\t\t</media>\n')
        outfile.write('\t</sequence>\n')
        outfile.write('</xmeml>\n')

    log.conwrite('')



'''audiotsm2/utils/cbuffer.py'''

import numpy as np

class CBuffer():
    def __init__(self, channels, max_length):
        self._data = np.zeros((channels, max_length), dtype=np.float32)
        self._channels = channels
        self._max_length = max_length

        self._offset = 0
        self._ready = 0
        self._length = 0

    def __repr__(self):
        return "CBuffer(offset={}, length={}, ready={}, data=\n{})".format(
            self._offset, self._length, self._ready, repr(self.to_array()))

    def add(self, buffer):
        """Adds a buffer element-wise to the CBuffer."""
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of "
                "channels")

        n = buffer.shape[1]
        if n > self._length:
            raise ValueError("not enough space remaining in CBuffer")

        # Compute the slice of data where the values will be added
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] += buffer[:, :n]
        else:
            end -= self._max_length
            self._data[:, start:] += buffer[:, :self._max_length - start]
            self._data[:, :end] += buffer[:, self._max_length - start:n]

    def divide(self, array):
        """
        Divides each channel of the CBuffer element-wise by the array.
        """
        n = len(array)
        if n > self._length:
            raise ValueError("not enough space remaining in the CBuffer")

        # Compute the slice of data where the values will be divided
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] /= array[:n]
        else:
            end -= self._max_length
            self._data[:, start:] /= array[:self._max_length - start]
            self._data[:, :end] /= array[self._max_length - start:n]

    @property
    def length(self):
        """The number of samples of each channel of the CBuffer."""
        return self._length

    def peek(self, buffer):
        """Reads as many samples from the :class:`CBuffer` as possible, without
        removing them from the :class:`CBuffer`, writes them to the ``buffer``,
        and returns the number of samples that were read.

        The samples need to be marked as ready to be read with the
        :func:`CBuffer.set_ready` method in order to be read. This is done
        automatically by the :func:`CBuffer.write` and
        :func:`CBuffer.read_from` methods.
        """
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError(
                "the two buffers should have the same number of channels")

        n = min(buffer.shape[1], self._ready)

        # Compute the slice of data the values will be read from
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            np.copyto(buffer[:, :n], self._data[:, start:end])
        else:
            end -= self._max_length
            np.copyto(buffer[:, :self._max_length - start],
                      self._data[:, start:])
            np.copyto(buffer[:, self._max_length - start:n],
                      self._data[:, :end])

        return n

    def read(self, buffer):
        """Reads as many samples from the CBuffer as possible, removes
        them from the CBuffer, writes them to the buffer, and
        returns the number of samples that were read.

        The samples need to be marked as ready to be read with the
        CBuffer.set_ready() method in order to be read. This is done
        automatically by the CBuffer.write() and
        CBuffer.read_from methods.
        """
        n = self.peek(buffer)
        self.remove(n)
        return n

    def read_from(self, reader):
        """Reads as many samples as possible from reader, writes them to
        the CBuffer, and returns the number of samples that were read.

        The written samples are marked as ready to be read.
        """

        # Compute the slice of data that will be written to
        start = (self._offset + self._length) % self._max_length
        end = start + self._max_length - self._length

        if end <= self._max_length:
            n = reader.read(self._data[:, start:end])
        else:
            # There is not enough space to copy the whole buffer, it has to be
            # split into two parts, one of which will be copied at the end of
            # _data, and the other at the beginning.
            end -= self._max_length

            n = reader.read(self._data[:, start:])
            n += reader.read(self._data[:, :end])

        self._length += n
        self._ready = self._length
        return n

    @property
    def ready(self):
        """The number of samples that can be read."""
        return self._ready

    @property
    def remaining_length(self):
        """The number of samples that can be added to the CBuffer."""
        return self._max_length - self._ready

    def remove(self, n):
        """
        Removes the first n samples of the CBuffer, preventing
        them to be read again, and leaving more space for new samples to be
        written.
        """
        if n >= self._length:
            n = self._length

        # Compute the slice of data that will be reset to 0
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] = 0
        else:
            end -= self._max_length
            self._data[:, start:] = 0
            self._data[:, :end] = 0

        self._offset += n
        self._offset %= self._max_length
        self._length -= n

        self._ready -= n
        if self._ready < 0:
            self._ready = 0

        return n

    def right_pad(self, n):
        """Add zeros at the end of the CBuffer.

        The added samples are not marked as ready to be read. The
        CBuffer.set_ready will need to be called in order to be able to
        read them.
        """
        if n > self._max_length - self._length:
            raise ValueError("not enough space remaining in CBuffer")

        self._length += n

    def set_ready(self, n):
        """Mark the next n samples as ready to be read."""
        if self._ready + n > self._length:
            raise ValueError("not enough samples to be marked as ready")

        self._ready += n

    def to_array(self):
        """Returns an array containing the same data as the CBuffer."""
        out = np.empty((self._channels, self._ready))
        self.peek(out)
        return out

    def write(self, buffer):
        """Writes as many samples from the buffer to the CBuffer
        as possible, and returns the number of samples that were read.

        The written samples are marked as ready to be read.
        """
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError(
                "the two buffers should have the same number of channels")

        n = min(buffer.shape[1], self._max_length - self._length)

        # Compute the slice of data that will be written to
        start = (self._offset + self._length) % self._max_length
        end = start + n

        if end <= self._max_length:
            np.copyto(self._data[:, start:end], buffer[:, :n])
        else:
            # There is not enough space to copy the whole buffer, it has to be
            # split into two parts, one of which will be copied at the end of
            # _data, and the other at the beginning.
            end -= self._max_length

            np.copyto(self._data[:, start:],
                      buffer[:, :self._max_length - start])
            np.copyto(self._data[:, :end],
                      buffer[:, self._max_length - start:n])

        self._length += n
        self._ready = self._length
        return n

    def write_to(self, writer):
        """Writes as many samples as possible to writer, deletes them from
        the CBuffer, and returns the number of samples that were
        written.

        The samples need to be marked as ready to be read with the
        CBuffer.set_ready method in order to be read. This is done
        automatically by the CBuffer.write and CBuffer.read_from methods.
        """

        # Compute the slice of data the values will be read from
        start = self._offset
        end = self._offset + self._ready

        if end <= self._max_length:
            n = writer.write(self._data[:, start:end])
        else:
            end -= self._max_length
            n = writer.write(self._data[:, start:])
            n += writer.write(self._data[:, :end])

        self.remove(n)
        return n



'''audiotsm2/io/array.py'''

import numpy as np

class ArrReader:
    pointer = 0

    def __init__(self, arr, channels, samplerate, samplewidth):
        self.samples = arr
        self._channels = channels
        self.samplerate = samplerate
        self.samplewidth = samplewidth

    @property
    def channels(self):
        return self._channels

    @property
    def empty(self):
        return self.samples.shape[0] <= self.pointer

    def read(self, buffer):
        if buffer.shape[0] != self.channels:
            raise ValueError(
                "the buffer should have the same number of channels as the ArrReader"
            )

        end = self.pointer + buffer.shape[1]
        frames = self.samples[self.pointer : end].T.astype(np.float32)
        n = frames.shape[1]
        np.copyto(buffer[:, :n], frames)
        del frames
        self.pointer = end
        return n

    def skip(self, n):
        pastPointer = self.pointer
        self.pointer += n
        return self.pointer - pastPointer

    def __enter__(self):
        return self

    def __exit__(self, _1, _2, _3):
        pass


class ArrWriter:
    pointer = 0

    def __init__(self, arr, channels, samplerate, samplewidth):
        self._channels = channels
        self.output = arr

    @property
    def channels(self):
        return self._channels

    def write(self, buffer):
        if buffer.shape[0] != self.channels:
            raise ValueError(
                "the buffer should have the same number of channels as the ArrWriter"
            )

        end = self.pointer + buffer.shape[1]
        changedBuffer = buffer.T.astype(np.int16)
        n = buffer.shape[1]
        self.output = np.concatenate((self.output, changedBuffer))
        self.pointer = end
        return n

    def __enter__(self):
        return self

    def __exit__(self, _1, _2, _3):
        pass



'''utils/func.py'''

"""
To prevent duplicate code being pasted between scripts, common functions should be
put here. No function should modify or create video/audio files on its own.
"""

def get_stdout(cmd):
    # type: (list[str]) -> str
    import subprocess
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)
    stdout, __ = process.communicate()
    return stdout.decode()

def clean_list(x, rm_chars):
    # type: (list, str) -> list
    new_list = []
    for item in x:
        for char in rm_chars:
            item = item.replace(char, '')
        new_list.append(item)
    return new_list

def aspect_ratio(width, height):
    # type: (int, int) -> tuple(2)
    if(height == 0):
        return None, None

    def gcd(a, b):
        while b:
            a, b = b, a % b
        return a

    c = gcd(width, height)
    return width // c, height // c

def get_new_length(chunks, speeds, fps):
    # type: (list[list[int]], list[float | int], float) -> float
    timeInFrames = 0
    for chunk in chunks:
        leng = chunk[1] - chunk[0]
        if(speeds[chunk[2]] < 99999):
            timeInFrames += leng * (1 / speeds[chunk[2]])
    return timeInFrames / fps

def human_readable_time(time_in_secs):
    # type(int | float) -> str
    units = 'seconds'
    if(time_in_secs >= 3600):
        time_in_secs = round(time_in_secs / 3600, 1)
        if(time_in_secs % 1 == 0):
            time_in_secs = round(time_in_secs)
        units = 'hours'
    if(time_in_secs >= 60):
        time_in_secs = round(time_in_secs / 60, 1)
        if(time_in_secs >= 10 or time_in_secs % 1 == 0):
            time_in_secs = round(time_in_secs)
        units = 'minutes'
    return '{} {}'.format(time_in_secs, units)

def open_with_system_default(path, log):
    from subprocess import call
    try:  # should work on Windows
        from os import startfile
        startfile(path)
    except ImportError:
        try:  # should work on MacOS and most Linux versions
            call(['open', path])
        except Exception:
            try: # should work on WSL2
                call(['cmd.exe', '/C', 'start', path])
            except Exception:
                try: # should work on various other Linux distros
                    call(['xdg-open', path])
                except Exception:
                    log.warning('Could not open output file.')

def fnone(val):
    # type: (Any) -> bool
    return val == 'none' or val == 'unset' or val is None

def append_filename(path, val):
    # type: (str, str) -> str
    import os.path
    root, ext = os.path.splitext(path)
    return root + val + ext

def set_output_name(path, inp_ext, making_data_file, args):
    import os.path
    root, ext = os.path.splitext(path)

    if(args.export_as_json):
        return root + '.json'
    if(args.export_to_final_cut_pro):
        return root + '.fcpxml'
    if(args.export_to_shotcut):
        return root + '.mlt'
    if(making_data_file):
        return root + '.xml'
    if(args.export_as_audio):
        return root + '_ALTERED.wav'
    if(ext == ''):
        if(inp_ext is None):
            return root
        return root + inp_ext
    return root + '_ALTERED' + ext



'''render/video.py'''

# Internal Libraries
import os.path
import subprocess

# From: github.com/PyAV-Org/PyAV/blob/main/av/video/frame.pyx
allowed_pix_fmt = {'yuv420p', 'yuvj420p', 'rgb24', 'bgr24', 'argb', 'rgba',
    'abgr', 'bgra', 'gray', 'gray8', 'rgb8', 'bgr8', 'pal8'}

def pix_fmt_allowed(pix_fmt):
    # type: (str) -> bool
    return pix_fmt in allowed_pix_fmt

class Wrapper:
    """
    Wrapper which only exposes the `read` method to avoid PyAV
    trying to use `seek`.
    From: github.com/PyAV-Org/PyAV/issues/578#issuecomment-621362337
    """

    name = "<wrapped>"

    def __init__(self, fh):
        self._fh = fh

    def read(self, buf_size):
        return self._fh.read(buf_size)


def render_av(ffmpeg, track, inp, args, chunks, speeds, fps, has_vfr, progress, effects,
    rules, temp, log):
    try:
        import av
    except ImportError:
        log.error("av python module not installed. Run 'pip install av'")

    total_frames = chunks[len(chunks) - 1][1]
    progress.start(total_frames, 'Creating new video')

    container = av.open(inp.path, 'r')
    pix_fmt = container.streams.video[track].pix_fmt

    if(has_vfr):
        # Create a cfr stream on stdout.
        cmd = ['-i', inp.path, '-map', f'0:v:{track}', '-vf', f'fps=fps={fps}', '-r',
            str(fps), '-vsync', '1', '-f', 'matroska']
        if(not pix_fmt_allowed(pix_fmt)):
            pix_fmt = 'yuv420p'
            cmd.extend(['-pix_fmt', pix_fmt])

        cmd.extend(['-vcodec', 'rawvideo', 'pipe:1'])

        wrapper = Wrapper(ffmpeg.Popen(cmd).stdout)
        container = av.open(wrapper, 'r')
    elif(not pix_fmt_allowed(pix_fmt)):
        pix_fmt = 'yuv420p'
        cmd = ['-i', inp.path, '-map', f'0:v:{track}', '-f', 'matroska', '-pix_fmt', pix_fmt,
            '-vcodec', 'rawvideo', 'pipe:1']
        wrapper = Wrapper(ffmpeg.Popen(cmd).stdout)
        container = av.open(wrapper, 'r')

    from auto_editor.utils.encoder import encoders
    from auto_editor.utils.video import get_vcodec, video_quality

    my_codec = get_vcodec(args, inp, rules)

    apply_video_later = True

    if(my_codec in encoders):
        apply_video_later = encoders[my_codec]['pix_fmt'].isdisjoint(allowed_pix_fmt)

    if(args.scale != 1):
        apply_video_later = False

    log.debug(f'pix_fmt: {pix_fmt}')
    log.debug(f'apply video quality settings now: {not apply_video_later}')

    video_stream = container.streams.video[track]
    video_stream.thread_type = 'AUTO'

    width = video_stream.width
    height = video_stream.height

    effects.add_var('fps', fps)
    effects.add_var('width', width)
    effects.add_var('height', height)
    effects.add_var('start', 0)
    effects.add_var('end', total_frames - 1)
    effects.add_var('centerX', width // 2)
    effects.add_var('centerY', height // 2)

    effects.resolve(args)

    spedup = os.path.join(temp, f'spedup{track}.mp4')

    cmd = ['-hide_banner', '-y', '-f', 'rawvideo', '-c:v', 'rawvideo',
        '-pix_fmt', pix_fmt, '-s', f'{width}*{height}', '-framerate', str(fps),
        '-i', '-', '-pix_fmt', pix_fmt]

    if(apply_video_later):
        cmd.extend(['-c:v', 'mpeg4', '-qscale:v', '1'])
    else:
        cmd = video_quality(cmd, args, inp, rules)

    cmd.append(spedup)

    process2 = ffmpeg.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)

    input_equavalent = 0.0
    output_equavalent = 0
    chunk = chunks.pop(0)

    try:
        for index, frame in enumerate(container.decode(video_stream)):
            if(len(chunks) > 0 and index-1 >= chunk[1]):
                chunk = chunks.pop(0)
                if(len(chunks) == 0 and speeds[chunk[2]] == 99999):
                    break

            if(speeds[chunk[2]] != 99999):
                input_equavalent += (1 / speeds[chunk[2]])

            while input_equavalent > output_equavalent:
                if(index in effects.sheet):
                    frame = effects.apply(index, frame, pix_fmt)

                in_bytes = frame.to_ndarray().tobytes()
                process2.stdin.write(in_bytes)
                output_equavalent += 1

            progress.tick(index)

        progress.end()
        process2.stdin.close()
        process2.wait()
    except (OSError, BrokenPipeError):
        progress.end()
        ffmpeg.run_check_errors(cmd, log, True)
        log.error('FFmpeg Error!')

    # Unfortunately, scaling has to be a concrete step.
    if(args.scale != 1):
        sped_input = os.path.join(temp, f'spedup{track}.mp4')
        spedup = os.path.join(temp, f'scale{track}.mp4')
        cmd = ['-i', sped_input, '-vf', 'scale=iw*{s}:ih*{s}'.format(s=args.scale),
            spedup]

        check_errors = ffmpeg.pipe(cmd)
        if('Error' in check_errors or 'failed' in check_errors):
            if('-allow_sw 1' in check_errors):
                # Add "-allow_sw 1" to command.
                cmd.insert(-1, '1')
                cmd.insert(-1, '-allow_sw')
            # Run again to show errors even if it might not work.
            ffmpeg.run(cmd)

    return 'video', spedup, apply_video_later

